# -*- coding: utf-8 -*-
"""
Created on Tue Oct  6 15:34:12 2015

@author: Numan Laanait -- nlaanait@gmail.com
"""

from __future__ import division, print_function, absolute_import
from skimage.feature import match_descriptors, register_translation
from skimage.measure import ransac
from skimage.transform import warp, SimilarityTransform
import warnings
import h5py
import numpy as np
import skimage.feature
import multiprocessing as mp


class ImageTransformation(object):
    #TODO: io operations and merging the 2 classes -Oleg
    # Oleg: reading, shaping data from h5.
    # Figure out storage of features and descriptors as well as reading.
    # Merge all methods from FeatureExtraction and  GeometricTransform.
    # Don't merge ancillary functions and transforms.

    pass

# TODO: Docstrings following numpy standard.

#### Functions
def pickle_keypoints(keypoints):
    """
    Function to pickle cv2.sift keypoint objects
    """
    kpArray = np.array([])
    for point in keypoints:
        kpArray = np.append(kpArray, [point.pt[1], point.pt[0]])
    kpArray = np.reshape(kpArray, (int(kpArray.size / 2), 2))
    return kpArray


# Class to do feature extraction. This is a wrapper on scikit-image and openCV feature extraction detectors.
# TODO: Add support for opencV or implement sift.
# TODO: Add io operations for extracted features.
# TODO: Memory checking, since some of the features are quite large.

class FeatureExtractorParallel(object):
    """
    This is an Object used to contain a data set and has methods to perform
    feature extraction on the data set that are detector based.
    Begin by loading a detector for features and a computer vision library.

    Parameters
    ----------
    detector_name : (string)
        name of detector.
    lib : (string)
        computer vision library to use (opencv or skimage)

        The following can be used for:
        lib = opencv: SIFT, ORB, SURF
        lib = skimage: ORB, BRIEF, CENSURE

    """

    def __init__(self, detector_name, lib):
        self.data = []
        self.lib = lib

        try:
            if self.lib == 'opencv':
                pass
                #                detector = cv2.__getattribute__(detector_name)
            elif self.lib == 'skimage':
                self.detector = skimage.feature.__getattribute__(detector_name)
        except AttributeError:
            print('Error: The Library does not contain the specified detector')

    def clearData(self):
        del self.data
        self.data = []

    def loadData(self, dataset):
        """
        This is a Method that loads h5 Dataset to be corrected.

        Parameters
        ----------
        dataset : h5py.Dataset
            Dataset to be corrected
        """
        if not isinstance(dataset, h5py.Dataset):
            warnings.warn('Error: Data must be an h5 Dataset object')
        else:
            self.data = dataset
            dim = int(np.sqrt(self.data.shape[-1]))
            self.data = self.data.reshape(-1, dim, dim)

    def getData(self):
        """
        This is a Method that returns the loaded h5 Dataset.

        """
        return self.data

    def getFeatures(self, **kwargs):
        """
        This is a Method that returns features (keypoints and descriptors)
        that are obtained by using the FeatureExtractor.Detector object.

        Parameters
        ----------
        processors : int, optional
                    Number of processors to use, default = 1.
        mask : boolean, optional, default False.
            Whether to use

        Returns
        -------
        keypts :
            keypoints
        descs :
            descriptors

        """
        detector = self.detector
        dset = self.data
        lib = self.lib
        processes = kwargs.get('processors', 1)
        mask = kwargs.get('mask', False)
        origin = kwargs.get('origin', [0, 0])
        winSize = kwargs.get('window_size', 0)

        if mask:
            def mask_func(x, winSize):
                x[origin[0] - winSize / 2: origin[0] + winSize / 2,
                origin[1] - winSize / 2: origin[1] + winSize / 2] = 2
                x = x - 1
                return x

            mask_ind = np.mask_indices(dset.shape[-1], mask_func, winSize)
            self.data = np.array([imp[mask_ind].reshape(winSize, winSize) for imp in dset])

        # detect and compute keypoints
        def detect(image):
            if lib == 'opencv':
                image = (image - image.mean()) / image.std()
                image = image.astype('uint8')
                k_obj, d_obj = detector.detectAndCompute(image, None)
                keypts, descs = pickle_keypoints(k_obj), pickle_keypoints(d_obj)

            elif lib == 'skimage':
                imp = (image - image.mean()) / np.std(image)
                imp[imp < 0] = 0
                imp.astype('float32')
                detector.detect_and_extract(imp)
                keypts, descs = detector.keypoints, detector.descriptors

            return keypts, descs

        # start pool of workers
        print('launching %i kernels...' % (processes))
        pool = mp.Pool(processes)
        tasks = [(imp) for imp in self.data]
        chunk = int(self.data.shape[0] / processes)
        jobs = pool.imap(detect, tasks, chunksize=chunk)

        # get keypoints and descriptors
        results = []
        print('Extracting features...')
        try:
            for j in jobs:
                results.append(j)
        except ValueError:
            warnings.warn('ValueError something about 2d-image. Probably some of the detector input params are wrong.')

        keypts = [itm[0].astype('int') for itm in results]
        desc = [itm[1] for itm in results]

        # close the pool
        print('Closing down the kernels... \n')
        pool.close()

        return keypts, desc


class FeatureExtractorSerial(object):
    """
    This is an Object used to contain a data set and has methods to perform
    feature extraction on the data set that are detector based.
    Begin by loading a detector for features and a computer vision library.

    Parameters
    ----------
    detector_name : (string)
        name of detector.
    lib : (string)
        computer vision library to use (opencv or skimage)

        The following can be used for:
        lib = opencv: SIFT, ORB, SURF
        lib = skimage: ORB, BRIEF, CENSURE

    """

    def __init__(self, detector_name, lib):
        self.data = []
        self.lib = lib

        try:
            if self.lib == 'opencv':
                pass
                #                detector = cv2.__getattribute__(detector_name)
            elif self.lib == 'skimage':
                self.detector = skimage.feature.__getattribute__(detector_name)
        except AttributeError:
            print('Error: The Library does not contain the specified detector')

    def clearData(self):
        del self.data
        self.data = []

    def loadData(self, dataset):
        """
        This is a Method that loads h5 Dataset to be corrected.

        Parameters
        ----------
        dataset : h5py.Dataset

        """
        if not isinstance(dataset, h5py.Dataset):
            warnings.warn('Error: Data must be an h5 Dataset object')
        else:
            self.data = dataset
            dim = int(np.sqrt(self.data.shape[-1]))
            self.data = self.data.reshape(-1, dim, dim)

    def getData(self):
        """
        This is a Method that returns the loaded h5 Dataset.
        """
        return self.data

    def getFeatures(self, **kwargs):
        """
        This is a Method that returns features (keypoints and descriptors)
        that are obtained by using the FeatureExtractor.Detector object.

        Parameters
        ----------
        mask : boolean, optional
            Whether to use, default False.

        Returns
        -------
        keypts :
            descriptors
        descs :
            keypoints

        """
        detector = self.detector
        dset = self.data
        lib = self.lib
        mask = kwargs.get('mask', False)
        origin = kwargs.get('origin', [0, 0])
        winSize = kwargs.get('window_size', 0)

        if mask:
            def mask_func(x, winSize):
                x[origin[0] - winSize / 2: origin[0] + winSize / 2,
                origin[1] - winSize / 2: origin[1] + winSize / 2] = 2
                x = x - 1
                return x

            mask_ind = np.mask_indices(dset.shape[-1], mask_func, winSize)
            self.data = np.array([imp[mask_ind].reshape(winSize, winSize) for imp in dset])

        # detect and compute keypoints
        def detect(image):
            if lib == 'opencv':
                image = (image - image.mean()) / image.std()
                image = image.astype('uint8')
                k_obj, d_obj = detector.detectAndCompute(image, None)
                keypts, descs = pickle_keypoints(k_obj), pickle_keypoints(d_obj)

            elif lib == 'skimage':
                imp = (image - image.mean()) / np.std(image)
                imp[imp < 0] = 0
                imp.astype('float32')
                detector.detect_and_extract(imp)
                keypts, descs = detector.keypoints, detector.descriptors

            return keypts, descs

        # start pool of workers
        results = [detect(imp) for imp in self.data]

        # get keypoints and descriptors
        keypts = [itm[0].astype('int') for itm in results]
        desc = [itm[1] for itm in results]

        return keypts, desc


#TODO: Docstrings following numpy standard.

# Functions
def euclidMatch(Matches, keypts1, keypts2, misalign):
    """
    Function that thresholds the matches, found from a comparison of
    their descriptors, by the maximum expected misalignment.
    """
    filteredMatches = np.array([])
    deltaX =(keypts1[Matches[:,0],:][:,0]-keypts2[Matches[:,1],:][:,0])**2
    deltaY =(keypts1[Matches[:,0],:][:,1]-keypts2[Matches[:,1],:][:,1])**2
    dist = np.apply_along_axis(np.sqrt, 0, deltaX + deltaY)
    filteredMatches = np.where(dist[:] < misalign, True, False)
    return filteredMatches


# function is taken as is from scikit-image.
def _center_and_normalize_points(points):
    """
    Center and normalize image points.

    The points are transformed in a two-step procedure that is expressed
    as a transformation matrix. The matrix of the resulting points is usually
    better conditioned than the matrix of the original points.

    Center the image points, such that the new coordinate system has its
    origin at the centroid of the image points.

    Normalize the image points, such that the mean distance from the points
    to the origin of the coordinate system is sqrt(2).

    Parameters
    ----------
    points : (N, 2) array
        The coordinates of the image points.

    Returns
    -------
    matrix : (3, 3) array
        The transformation matrix to obtain the new points.
    new_points : (N, 2) array
        The transformed image points.

    """

    centroid = np.mean(points, axis=0)

    rms = np.sqrt(np.sum((points - centroid) ** 2) / points.shape[0])

    norm_factor = np.sqrt(2) / rms

    matrix = np.array([[norm_factor, 0, -norm_factor * centroid[0]],
                       [0, norm_factor, -norm_factor * centroid[1]],
                       [0, 0, 1]])

    pointsh = np.row_stack([points.T, np.ones((points.shape[0]),)])

    new_pointsh = np.dot(matrix, pointsh).T

    new_points = new_pointsh[:, :2]
    new_points[:, 0] /= new_pointsh[:, 2]
    new_points[:, 1] /= new_pointsh[:, 2]

    return matrix, new_points


class TranslationTransform(object):
    """
    2D translation using homogeneous representation:

    The transformation matrix is:
        [[1  1  tX]
         [1  1  tY]
         [0  0  1]]
         X: translation of x-axis.
         Y: translation of y-axis.

    Parameters
    ----------
    translation : tuple
        (tX, tY)

    Attributes
    ----------
    params : (3, 3) array
        Homogeneous transformation matrix.

    """

    def __init__(self, matrix = None, translation = None):
        params = translation

        if params and matrix is not None:
            raise ValueError("You cannot specify the transformation matrix and"
                             " the implicit parameters at the same time.")
        elif matrix is not None:
            if matrix.shape != (3, 3):
                raise ValueError("Invalid shape of transformation matrix.")
            self.params = matrix

        elif params:
            if translation is None:
                translation = (0., 0.)

            self.params = np.array([
                [1., 0., 0.],
                [0., 1., 0.],
                [0., 0., 1.]
                ], dtype = 'float32')
            self.params[0:2, 2] = translation
        else:
            # default to an identity transform
            self.params = np.eye(3)

    def estimate(self, src, dst):
     #evaluate transformation matrix from src, dst
     # coordinates
        try:
            xs = src[:, 0][0]
            ys = src[:, 1][1]
            xd = dst[:, 0][0]
            yd = dst[:, 1][1]
            S = np.array([[1., 0., xd-xs],
                          [0., 1., yd-ys],
                          [0., 0., 1.]
                          ],dtype = 'float32')
            self.params = S
            return True
        except IndexError:
            return False

    @property
    def _inv_matrix(self):
        inv_matrix = self.params
        inv_matrix[0:2,2] = - inv_matrix[0:2,2]
        return inv_matrix

    def _apply_mat(self, coords, matrix):
        coords = np.array(coords, copy=False, ndmin=2)

        x, y = np.transpose(coords)
        src = np.vstack((x, y, np.ones_like(x)))
        dst = np.dot(src.transpose(), matrix.transpose())

        # rescale to homogeneous coordinates
        dst[:, 0] /= dst[:, 2]
        dst[:, 1] /= dst[:, 2]

        return dst[:, :2]

    def __call__(self, coords):
        return self._apply_mat(coords, self.params)

    def inverse(self, coords):
        """ Apply inverse transformation.

        Parameters
        ----------
        coords : (N, 2) array
            Source coordinates.

        Returns
        -------
        coords : (N, 2) array
            Transformed coordinates.

        """
        return self._apply_mat(coords, self._inv_matrix)

    def residuals(self, src, dst):
        """
        Determine residuals of transformed destination coordinates.

        For each transformed source coordinate the euclidean distance to the
        respective destination coordinate is determined.

        Parameters
        ----------
        src : (N, 2) array
            Source coordinates.
        dst : (N, 2) array
            Destination coordinates.

        Returns
        -------
        residuals : (N, ) array
            Residual for coordinate.

        """

        return np.sqrt(np.sum((self(src) - dst)**2, axis=1))

    @property
    def translation(self):
        return self.params[0:2, 2]


class RigidTransform(object):
    """
    2D translation using homogeneous representation:

    The transformation matrix is:
        [[cos(theta)  -sin(theta)  tX]
         [sin(theta)  cos(theta)   tY]
         [0             0           1]]
         X: translation along x-axis.
         Y: translation along y-axis.
         theta: rotation angle in radians.

    Parameters
    ----------
    translation : tuple
        (tX, tY)
    rotation : float
        in radians.

    Attributes
    ----------
    params : (3, 3) array
        Homogeneous transformation matrix.

    """

    def __init__(self, matrix = None, rotation = None, translation = None):
        params = any(param is not None
                     for param in (rotation, translation))

        if params and matrix is not None:
            raise ValueError("You cannot specify the transformation matrix and"
                             " the implicit parameters at the same time.")
        elif matrix is not None:
            if matrix.shape != (3, 3):
                raise ValueError("Invalid shape of transformation matrix.")
            self.params = matrix

        elif params:
            if translation is None:
                translation = (0, 0)
            if rotation is None:
                rotation = 0

            self.params = np.array([
                [np.cos(rotation), - np.sin(rotation), 0],
                [np.sin(rotation),   np.cos(rotation), 0],
                [                 0,                    0, 1]
            ])

            self.params[0:2, 2] = translation
        else:
            # default to an identity transform
            self.params = np.eye(3)

    def estimate(self, src, dst):
        """
        Set the transformation matrix with the explicit parameters.

        You can determine the over-, well- and under-determined parameters
        with the total least-squares method.

        Number of source and destination coordinates must match.

        The transformation is defined as::

            X = a0 * x - b0 * y + a1
            Y = b0 * x + a0 * y + b1

        These equations can be transformed to the following form::

            0 = a0 * x - b0 * y + a1 - X
            0 = b0 * x + a0 * y + b1 - Y

        which exist for each set of corresponding points, so we have a set of
        N * 2 equations. The coefficients appear linearly so we can write
        A x = 0, where::

            A   = [[x 1 -y 0 -X]
                   [y 0  x 1 -Y]
                    ...
                    ...
                  ]
            x.T = [a0 a1 b0 b1 c3]

        In case of total least-squares the solution of this homogeneous system
        of equations is the right singular vector of A which corresponds to the
        smallest singular value normed by the coefficient c3.

        Parameters
        ----------
        src : (N, 2) array
            Source coordinates.
        dst : (N, 2) array
            Destination coordinates.

        Returns
        -------
        success : bool
            True, if model estimation succeeds.

        """

        try:
            src_matrix, src = _center_and_normalize_points(src)
            dst_matrix, dst = _center_and_normalize_points(dst)
        except ZeroDivisionError:
            self.params = np.nan * np.empty((3, 3))
            return False

        xs = src[:, 0]
        ys = src[:, 1]
        xd = dst[:, 0]
        yd = dst[:, 1]
        rows = src.shape[0]

        # params: a0, a1, b0, b1
        A = np.zeros((rows * 2, 5))
        A[:rows, 0] = xs
        A[:rows, 2] = - ys
        A[:rows, 1] = 1
        A[rows:, 2] = xs
        A[rows:, 0] = ys
        A[rows:, 3] = 1
        A[:rows, 4] = xd
        A[rows:, 4] = yd

        _, _, V = np.linalg.svd(A)

        # solution is right singular vector that corresponds to smallest
        # singular value
        a0, a1, b0, b1 = - V[-1, :-1] / V[-1, -1]

        S = np.array([[a0, -b0, a1],
                      [b0,  a0, b1],
                      [ 0,   0,  1]])

        # De-center and de-normalize
        S = np.dot(np.linalg.inv(dst_matrix), np.dot(S, src_matrix))

        self.params = S

        return True

    def _apply_mat(self, coords, matrix):
        coords = np.array(coords, copy=False, ndmin=2)

        x, y = np.transpose(coords)
        src = np.vstack((x, y, np.ones_like(x)))
        dst = np.dot(src.transpose(), matrix.transpose())

        # rescale to homogeneous coordinates
        dst[:, 0] /= dst[:, 2]
        dst[:, 1] /= dst[:, 2]

        return dst[:, :2]

    def __call__(self, coords):
        return self._apply_mat(coords, self.params)

    def inverse(self, coords):
        """
        Apply inverse transformation.

        Parameters
        ----------
        coords : (N, 2) array
            Source coordinates.

        Returns
        -------
        coords : (N, 2) array
            Transformed coordinates.

        """
        return self._apply_mat(coords, self._inv_matrix)

    def residuals(self, src, dst):
        """
        Determine residuals of transformed destination coordinates.

        For each transformed source coordinate the euclidean distance to the
        respective destination coordinate is determined.

        Parameters
        ----------
        src : (N, 2) array
            Source coordinates.
        dst : (N, 2) array
            Destination coordinates.

        Returns
        -------
        residuals : (N, ) array
            Residual for coordinate.

        """

        return np.sqrt(np.sum((self(src) - dst)**2, axis=1))


    @property
    def _inv_matrix(self):
        return np.linalg.inv(self.params)

    @property
    def rotation(self):
        return np.atan2(self.params[1, 0], self.params[1, 1])

    @property
    def translation(self):
        return self.params[0:2, 2]



# Class to do geometric transformations. This is a wrapper on scikit-image functionality.
# TODO: io operations for features and optical geometric transformations.

class geoTransformerParallel(object):
    """
    This object contains methods to perform geometric transformations on
    a sequence of images. Some of the capabilities are:
    + Homography by feature extraction.
    + Intensity-based image registration.
    + Projection Correction.
    """

    def __init__(self):
        self.__init__
        self.data = []
        self.features = []

    def clearData(self):
        """
        This is a Method to clear the data from the object.
        """
        del self.data
        self.data = []

    def loadData(self, dataset):
        """
        This is a Method that loads h5 Dataset to be corrected.

        Parameters
        ----------
        input: h5py.dataset

        """
        if not isinstance(dataset, h5py.Dataset):
            warnings.warn( 'Error: Data must be an h5 Dataset object'   )
        else:
            self.data = dataset
            dim = int(np.sqrt(self.data.shape[-1]))
            self.data = self.data.reshape(-1,dim,dim)

    def loadFeatures(self, features):
        """
        This is a Method that loads features to be used for homography etc ...

        Parameters
        ----------
        features : tuple
            [keypoints, descriptors]
            These can come from FeatureExtractor.getFeatures() or elsewhere.

            The format is :
            keypoints = [np.ndarray([y_position, x_position])]
            descriptors = [np.ndarray()]

        """
        self.features = features

    def matchFeatures(self, **kwargs):
        """
        This is a Method that computes similarity between keypoints based on their
        descriptors. Currently only skimage.feature.match_descriptors is implemented.
        In the future will need to add opencv2.matchers.

        Parameters
        ----------
        processors: int, optional
                Number of processors to use, default = 1.
        maximum_distance: int, optional
                maximum_distance (int) of misalignment, default = infinity.
                Used to filter the matches before optimizing the transformation.

        Returns
        -------
            Matches

        """
        desc = self.features[-1]
        keypts = self.features[0]
        processes = kwargs.get('processors', 1)
        maxDis = kwargs.get('maximum_distance', np.infty)

        def match(desc):
            desc1, desc2 = desc[0], desc[1]
            matches = match_descriptors(desc1, desc2, cross_check=True)
            return matches

        # start pool of workers
        pool = mp.Pool(processes)
        print('launching %i kernels...'%(processes))

        tasks = [ (desc1, desc2) for desc1, desc2 in zip(desc[:],desc[1:]) ]
        chunk = int(len(desc)/processes)
        jobs = pool.imap(match, tasks, chunksize = chunk)

        # get matches
        print('Extracting Matches From the Descriptors...')

        matches =[]
        for j in jobs:
            matches.append(j)

        # close the pool
        print('Closing down the kernels...\n')
        pool.close()

        # impose maximum_distance misalignment constraints on matches
        filt_matches = []
        for match, key1, key2 in zip(matches, keypts[:],keypts[1:]):
            filteredMask = euclidMatch(match, key1, key2, maxDis)
            filt_matches.append(match[filteredMask])


        return matches, filt_matches

    def findTransformation(self, transform, matches, processes, **kwargs):
        """
        This is a Method that finds the optimal transformation between two images
        given matching features using a random sample consensus.

        Parameters
        ----------
        transform: skimage.transform object
        matches : list
            matches found through match_features method.
        processes : int
            Number of processors to use.
        **kwargs are passed to skimage.transform.ransac

        Returns
        -------
        Transformations

        """

        keypts = self.features[0]

        def optimization(Pts):
            robustTrans, inliers = ransac((Pts[0], Pts[1]), transform, **kwargs)
            output = [robustTrans, inliers]
            return output

         # start pool of workers
        print('launching %i kernels...'%(processes))
        pool = mp.Pool(processes)
        tasks = [ (key1[match[:, 0]], key2[match[:, 1]])
                    for match, key1, key2 in zip(matches,keypts[:],keypts[1:]) ]
        chunk = int(len(keypts)/processes)
        jobs = pool.imap(optimization, tasks, chunksize = chunk)

        # get Transforms and inlier matches
        transforms, trueMatches =[], []
        print('Extracting Inlier Matches with RANSAC...')
        try:
            for j in jobs:
                transforms.append(j[0])
                trueMatches.append(j[1])
        except np.linalg.LinAlgError:
            pass

        # close the pool
        pool.close()
        print('Closing down the kernels...\n')

        return transforms, trueMatches

    #TODO: Need parallel version for transforming stack of images.
    def applyTransformation(self, transforms, **kwargs):
        """
        This is the method that takes the list of transformation found by findTransformation
        and applies them to the data set.

        Parameters
        ----------
        transforms: (list of skimage.GeoemetricTransform objects).
             The objects must be inititated with the desired parameters.
        transformation : string, optional.
             The type of geometric transformation to use (i.e. translation, rigid, etc..)
             Currently, only translation is implemented.
             default, translation.
        origin : int, optional
             The position in the data to take as origin, i.e. don't transform.
             default, center image in the stack.
        processors : int, optional
            Number of processors to use, default = 1.
            Currently,only one processor is used.

        Returns
        -------
        Transformed images, transformations

        """
        dic = ['processors','origin','transformation']
        for key in kwargs.keys():
            if key not in dic:
                print('%s is not a parameter of this function' %(str(key)))

        processes = kwargs.get('processors', 1)
        origin = kwargs.get('origin', int(self.data.shape[0]/2))
        transformation = kwargs.get('transformation','translation')

        dset = self.data
        # For now restricting this to just translation... Straightforward to generalize to other transform objects.
        if transformation == 'translation':

            YTrans = np.array([trans.translation[0] for trans in transforms])
            XTrans = np.array([trans.translation[1] for trans in transforms])
            chainL = []
            for y, x in zip(range(0,YTrans.size+1), range(0,XTrans.size+1)):
                if y < origin:
                    ychain = -np.sum(YTrans[y:origin])
                    xchain = -np.sum(XTrans[x:origin])

                elif y > origin:
                    ychain = np.sum(YTrans[origin:y])
                    xchain = np.sum(XTrans[origin:x])
                else:
                    ychain = 0
                    xchain = 0

                chainL.append([xchain, ychain])

            chainTransforms = []
            for params in  chainL:
                T = TranslationTransform(translation = params)
                chainTransforms.append(T)

        # Just need a single function that does boths
        if transformation == 'rotation':

            rotTrans = np.array([trans.rotation for trans in transforms])
            YTrans = np.array([trans.translation[0] for trans in transforms])
            XTrans = np.array([trans.translation[1] for trans in transforms])
            chainL = []
            for x in range(0,rotTrans.size+1):
                if x < origin:
                    rotchain = -np.sum(rotTrans[x:origin])
                    ychain = -np.sum(YTrans[x:origin])
                    xchain = -np.sum(XTrans[x:origin])

                elif x > origin:
                    rotchain = np.sum(rotTrans[origin:x])
                    ychain = np.sum(YTrans[origin:x])
                    xchain = np.sum(XTrans[origin:x])
                else:
                    rotchain = 0
                    ychain = 0
                    xchain = 0

                chainL.append([rotchain, xchain, ychain])

            chainTransforms = []
            for params in  chainL:
                T = SimilarityTransform(scale = 1.0, rotation = np.deg2rad(params[0]), translation = (params[1],params[2]))
#                T = SimilarityTransform(rotation = params, translation = (0,0))
                chainTransforms.append(T)

        # Use the chain transformations to transform the dataset
        output_shape = dset[0].shape
#        output_shape = (2048, 2048)
        def warping(datum):
            imp, transform  = datum[0], datum[1]
            transimp = warp(imp, inverse_map= transform, output_shape = output_shape,
                            cval = 0, preserve_range = True)
            return transimp

#          #start pool of workers
#         #somehow wrap function crashes when run in parallel! run sequentially for now.
#        pool = mp.Pool(processes)
#        print('launching %i kernels...'%(processes))
#        tasks = [ (imp, transform) for imp, transform in zip(dset, chainTransforms) ]
#        chunk = int(dset.shape[0]/processes)
#        jobs = pool.imap(warping, tasks, chunksize = 1)
#        #close the pool
#        pool.close()
#        print('Closing down the kernels... \n')
#
        # get transformed images and pack into 3d np.ndarray
        print('Transforming Images...')
        transImages = np.copy(dset[:])

        for imp, transform, itm in zip( transImages, chainTransforms, range(0,transImages.shape[0])):
            transimp = warping([imp, transform])
            transImages[itm] = transimp
            print('Image #%i'%(itm))


        return transImages, chainTransforms

    def correlationTransformation(self, **kwargs):
        """
        Uses Cross-correlation to find a translation between 2 images.

        Parameters
        ----------
        Processors: int, optional
            Number of processors to use, default = 1.

        Returns
        -------
        Transformations.

        """

        processes = kwargs.get('processors', 1)

        pool = mp.Pool(processes)
        print('launching %i kernels...'%(processes))

        def register(images):
            imp1, imp2 = images[0], images[1]
            shifts, _, _ = register_translation(imp1,imp2)
            return shifts

        dim = int(np.sqrt(self.data.shape[-1]))
        tasks = [ (imp1, imp2)
                    for imp1, imp2 in zip(self.data[:], self.data[1:]) ]

        chunk = int((self.data.shape[0] - 1)/processes)
        jobs = pool.imap(register, tasks, chunksize = chunk)

        # get Transforms and inlier matches
        results = []
        print('Extracting Translations')
        try:
            for j in jobs:
                results.append(j)
        except:
            warnings.warn('Skipped Some Entry... dunno why!!')

        # close the pool
        pool.close()

        return results


class geoTransformerSerial(object):
    """
    This object contains methods to perform geometric transformations on
    a sequence of images. Some of the capabilities are:
    + Homography by feature extraction.
    + Intensity-based image registration.
    + Projection Correction.
    """

    def __init__(self):
        self.__init__
        self.data = []
        self.features = []

    def clearData(self):
        """
        This is a Method to clear the data from the object.
        """
        del self.data
        self.data = []

    def loadData(self, dataset):
        """
        This is a Method that loads h5 Dataset to be corrected.

        Parameters
        ----------
        dataset : h5py.dataset

        """
        if not isinstance(dataset, h5py.Dataset):
            warnings.warn( 'Error: Data must be an h5 Dataset object'   )
        else:
            self.data = dataset
            dim = int(np.sqrt(self.data.shape[-1]))
            self.data = self.data.reshape(-1,dim,dim)

    def loadFeatures(self, features):
        """
        This is a Method that loads features to be used for homography etc ...

        Parameters
        ----------
        features : tuple
            [keypoints, descriptors]
            These can come from FeatureExtractor.getFeatures() or elsewhere.

            The format is :
            keypoints = [np.ndarray([y_position, x_position])]
            descriptors = [np.ndarray()]

        """
        self.features = features

    def matchFeatures(self, **kwargs):
        """
        This is a Method that computes similarity between keypoints based on their
        descriptors. Currently only skimage.feature.match_descriptors is implemented.
        In the future will need to add opencv2.matchers.

        Parameters
        ----------
        maximum_distance: int, optional
            maximum_distance (int) of misalignment, default = infinity.
            Used to filter the matches before optimizing the transformation.

        Returns
        -------
        Matches.

        """
        desc = self.features[-1]
        keypts = self.features[0]
        maxDis = kwargs.get('maximum_distance', np.infty)
        processes = kwargs.get('processes', 2)


        def match(desc):
            desc1, desc2 = desc[0], desc[1]
            matches = match_descriptors(desc1, desc2, cross_check=True)
            return matches

        # start pool of workers
        pool = mp.Pool(processes)
        print('launching %i kernels...'%(processes))

        tasks = [ (desc1, desc2) for desc1, desc2 in zip(desc[:],desc[1:]) ]
        chunk = int(len(desc)/processes)
        jobs = pool.imap(match, tasks, chunksize = chunk)

        # get matches
        print('Extracting Matches From the Descriptors...')

        matches =[]
        for j in jobs:
            matches.append(j)

        # close the pool
        print('Closing down the kernels...\n')
        pool.close()

        # impose maximum_distance misalignment constraints on matches
        filt_matches = []
        for match, key1, key2 in zip(matches, keypts[:],keypts[1:]):
            filteredMask = euclidMatch(match, key1, key2, maxDis)
            filt_matches.append(match[filteredMask])


        return matches, filt_matches

    #TODO: Need Better Error Handling.
    def findTransformation(self, transform, matches, processes, **kwargs):
        """
        This is a Method that finds the optimal transformation between two images
        given matching features using a random sample consensus.

        Parameters
        ----------
        transform : skimage.transform object
        matches : list
            matches found through match_features method.
        processors : int
            Number of processors to use.
        **kwargs are passed to skimage.transform.ransac

        Returns
        -------
        Transformations.

        """

        keypts = self.features[0]

        def optimization(Pts):
            robustTrans, inliers = ransac((Pts[0], Pts[1]), transform, **kwargs)
            output = [robustTrans, inliers]
            return output

        results = [optimization(key1[match[:, 0]], key2[match[:, 1]])
                    for match, key1, key2 in zip(matches,keypts[:],keypts[1:])]

        # get Transforms and inlier matches
        transforms, trueMatches =[], []
        print('Extracting Inlier Matches with RANSAC...')
        try:
            for res in results:
                transforms.append(res[0])
                trueMatches.append(res[1])
        except np.linalg.LinAlgError:
            print('Error: Inverse of the transformation failed!!!')

        return transforms, trueMatches

    def applyTransformation(self, transforms, **kwargs):
        """
        This is the method that takes the list of transformation found by findTransformation
        and applies them to the data set.

        Parameters
         transforms: (list of skimage.GeoemetricTransform objects).
                 The objects must be inititated with the desired parameters.
         transformation: string, optional.
                 The type of geometric transformation to use (i.e. translation, rigid, etc..)
                 Currently, only translation is implemented.
                 default, translation.
         origin: int, optional
                 The position in the data to take as origin, i.e. don't transform.
                 default, center image in the stack.

        Returns
        -------
        Transformed images, transformations

        """
        dic = ['processors','origin','transformation']
        for key in kwargs.keys():
            if key not in dic:
                print('%s is not a parameter of this function' %(str(key)))

        processes = kwargs.get('processors', 1)
        origin = kwargs.get('origin', int(self.data.shape[0]/2))
        transformation = kwargs.get('transformation','translation')

        dset = self.data
        # For now restricting this to just translation... Straightforward to generalize to other transform objects.
        if transformation == 'translation':

            YTrans = np.array([trans.translation[0] for trans in transforms])
            XTrans = np.array([trans.translation[1] for trans in transforms])
            chainL = []
            for y, x in zip(range(0,YTrans.size+1), range(0,XTrans.size+1)):
                if y < origin:
                    ychain = -np.sum(YTrans[y:origin])
                    xchain = -np.sum(XTrans[x:origin])

                elif y > origin:
                    ychain = np.sum(YTrans[origin:y])
                    xchain = np.sum(XTrans[origin:x])
                else:
                    ychain = 0
                    xchain = 0

                chainL.append([xchain, ychain])

            chainTransforms = []
            for params in  chainL:
                T = TranslationTransform(translation = params)
                chainTransforms.append(T)

        # Just need a single function that does boths
        if transformation == 'rotation':

            rotTrans = np.array([trans.rotation for trans in transforms])
            YTrans = np.array([trans.translation[0] for trans in transforms])
            XTrans = np.array([trans.translation[1] for trans in transforms])
            chainL = []
            for x in range(0,rotTrans.size+1):
                if x < origin:
                    rotchain = -np.sum(rotTrans[x:origin])
                    ychain = -np.sum(YTrans[x:origin])
                    xchain = -np.sum(XTrans[x:origin])

                elif x > origin:
                    rotchain = np.sum(rotTrans[origin:x])
                    ychain = np.sum(YTrans[origin:x])
                    xchain = np.sum(XTrans[origin:x])
                else:
                    rotchain = 0
                    ychain = 0
                    xchain = 0

                chainL.append([rotchain, xchain, ychain])

            chainTransforms = []
            for params in  chainL:
                T = SimilarityTransform(scale = 1.0, rotation = np.deg2rad(params[0]), translation = (params[1],params[2]))
#                T = SimilarityTransform(rotation = params, translation = (0,0))
                chainTransforms.append(T)

        # Use the chain transformations to transform the dataset
        output_shape = dset[0].shape
        def warping(datum):
            imp, transform  = datum[0], datum[1]
            transimp = warp(imp, inverse_map= transform, output_shape = output_shape,
                            cval = 0, preserve_range = True)
            return transimp

        # get transformed images and pack into 3d np.ndarray
        print('Transforming Images...')
        transImages = np.copy(dset[:])

        for imp, transform, itm in zip( transImages, chainTransforms, range(0,transImages.shape[0])):
            transimp = warping([imp, transform])
            transImages[itm] = transimp
            print('Image #%i'%(itm))


        return transImages, chainTransforms

    def correlationTransformation(self, **kwargs):
        """
        Uses Cross-correlation to find a translation between 2 images.

        Parameters
        ----------
        Processors: int, optional
            Number of processors to use, default = 1.

        Returns
        -------
        Transformations.

        """

        processes = kwargs.get('processors', 1)

        pool = mp.Pool(processes)
        print('launching %i kernels...'%(processes))

        def register(images):
            imp1, imp2 = images[0], images[1]
            shifts, _, _ = register_translation(imp1,imp2)
            return shifts

        results = [register((imp1, imp2))
                    for imp1, imp2 in zip(self.data[:], self.data[1:])]

        return results




# -*- coding: utf-8 -*-
"""
Created on Thu Jan 05 13:29:12 2017

@author: Suhas Somnath, Chris R. Smith
"""
from __future__ import division, print_function, absolute_import, unicode_literals
import os
from warnings import warn
import numpy as np
from matplotlib import pyplot as plt
import ipywidgets as widgets
from IPython.display import display

from .plot_utils import plot_loops, plot_map_stack, cmap_jet_white_center, plot_map
from ..io.hdf_utils import reshape_to_Ndims, getAuxData, get_sort_order, get_dimensionality, get_attr
from ..analysis.utils.be_loop import loop_fit_function


def visualize_sho_results(h5_main, save_plots=True, show_plots=True):
    """
    Plots some loops, amplitude, phase maps for BE-Line and BEPS datasets.\n
    Note: The file MUST contain SHO fit gusses at the very least

    Parameters
    ----------
    h5_main : HDF5 Dataset
        dataset to be plotted
    save_plots : (Optional) Boolean
        Whether or not to save plots to files in the same directory as the h5 file
    show_plots : (Optional) Boolean
        Whether or not to display the plots on the screen

    Returns
    -------
    None
    """

    def __plot_loops_maps(ac_vec, resp_mat, grp_name, win_title, spec_var_title, meas_var_title, save_plots,
                          folder_path, basename, num_rows, num_cols):
        plt_title = grp_name + '_' + win_title + '_Loops'
        fig, ax = plot_loops(ac_vec, resp_mat, evenly_spaced=True, plots_on_side=5, use_rainbow_plots=False,
                             x_label=spec_var_title, y_label=meas_var_title, subtitles='Position', title=plt_title)
        if save_plots:
            fig.savefig(os.path.join(folder_path, basename + '_' + plt_title + '.png'), format='png', dpi=300)

        plt_title = grp_name + '_' + win_title + '_Snaps'
        fig, axes = plot_map_stack(resp_mat.reshape(num_rows, num_cols, resp_mat.shape[1]),
                                   color_bar_mode="each", evenly_spaced=True, title='UDVS Step #',
                                   heading=plt_title, cmap=plt.cm.viridis)
        if save_plots:
            fig.savefig(os.path.join(folder_path, basename + '_' + plt_title + '.png'), format='png', dpi=300)

    plt_path = None

    print('Creating plots of SHO Results from {}.'.format(h5_main.name))

    h5_file = h5_main.file

    expt_type = get_attr(h5_file, 'data_type')
    if expt_type not in ['BEPSData', 'BELineData']:
        warn('Unsupported data format')
        return
    isBEPS = expt_type == 'BEPSData'

    (folder_path, basename) = os.path.split(h5_file.filename)
    basename, _ = os.path.splitext(basename)

    sho_grp = h5_main.parent

    chan_grp = h5_file['/'.join(sho_grp.name[1:].split('/')[:2])]

    grp_name = '_'.join(chan_grp.name[1:].split('/'))
    grp_name = '_'.join([grp_name, sho_grp.name.split('/')[-1].split('-')[0], h5_main.name.split('/')[-1]])

    try:
        h5_pos = h5_file[get_attr(h5_main, 'Position_Indices')]
    except KeyError:
        print('No Position_Indices found as attribute of {}'.format(h5_main.name))
        print('Rows and columns will be calculated from dataset shape.')
        num_rows = int(np.floor((np.sqrt(h5_main.shape[0]))))
        num_cols = int(np.reshape(h5_main, [num_rows, -1, h5_main.shape[1]]).shape[1])
    else:
        num_rows = len(np.unique(h5_pos[:, 0]))
        num_cols = len(np.unique(h5_pos[:, 1]))

    try:
        h5_spec_vals = h5_file[get_attr(h5_main, 'Spectroscopic_Values')]
    # except KeyError:
    #     warn('No Spectrosocpic Datasets found as attribute of {}'.format(h5_main.name))
    #     raise
    except:
        raise

    # Assume that there's enough memory to load all the guesses into memory
    amp_mat = h5_main['Amplitude [V]'] * 1000  # convert to mV ahead of time
    freq_mat = h5_main['Frequency [Hz]'] / 1000
    q_mat = h5_main['Quality Factor']
    phase_mat = h5_main['Phase [rad]']
    rsqr_mat = h5_main['R2 Criterion']

    if isBEPS:
        meas_type = chan_grp.parent.attrs['VS_mode']
        # basically 3 kinds for now - DC/current, AC, UDVS - lets ignore this
        if meas_type == 'load user defined VS Wave from file':
            warn('Not handling custom experiments for now')
            # h5_file.close()
            return

        # Plot amplitude and phase maps at one or more UDVS steps

        if meas_type == 'AC modulation mode with time reversal':
            center = int(h5_spec_vals.shape[1] * 0.5)
            ac_vec = np.squeeze(h5_spec_vals[h5_spec_vals.attrs['AC_Amplitude']][:, 0:center])

            forw_resp = np.squeeze(amp_mat[:, slice(0, center)])
            rev_resp = np.squeeze(amp_mat[:, slice(center, None)])

            for win_title, resp_mat in zip(['Forward', 'Reverse'], [forw_resp, rev_resp]):
                __plot_loops_maps(ac_vec, resp_mat, grp_name, win_title, 'AC Amplitude', 'Amplitude', save_plots,
                                  folder_path, basename, num_rows, num_cols)
        else:
            # plot loops at a few locations
            dc_vec = np.squeeze(h5_spec_vals[h5_spec_vals.attrs['DC_Offset']])
            if chan_grp.parent.attrs['VS_measure_in_field_loops'] == 'in and out-of-field':

                dc_vec = np.squeeze(dc_vec[slice(0, None, 2)])

                in_phase = np.squeeze(phase_mat[:, slice(0, None, 2)])
                in_amp = np.squeeze(amp_mat[:, slice(0, None, 2)])
                out_phase = np.squeeze(phase_mat[:, slice(1, None, 2)])
                out_amp = np.squeeze(amp_mat[:, slice(1, None, 2)])

                for win_title, resp_mat in zip(['In_Field', 'Out_of_Field'], [in_phase * in_amp, out_phase * out_amp]):
                    __plot_loops_maps(dc_vec, resp_mat, grp_name, win_title, 'DC Bias', 'Piezoresponse (a.u.)',
                                      save_plots, folder_path, basename, num_rows, num_cols)
            else:
                __plot_loops_maps(dc_vec, phase_mat * amp_mat, grp_name, '', 'DC Bias', 'Piezoresponse (a.u.)',
                                  save_plots, folder_path, basename, num_rows, num_cols)

    else:  # BE-Line can only visualize the amplitude and phase maps:
        amp_mat = amp_mat.reshape(num_rows, num_cols)
        freq_mat = freq_mat.reshape(num_rows, num_cols)
        q_mat = q_mat.reshape(num_rows, num_cols)
        phase_mat = phase_mat.reshape(num_rows, num_cols)
        rsqr_mat = rsqr_mat.reshape(num_rows, num_cols)

        fig_ms, ax_ms = plot_map_stack(np.dstack((amp_mat, freq_mat, q_mat, phase_mat, rsqr_mat)),
                                       num_comps=5, color_bar_mode='each', heading=grp_name,
                                       title=['Amplitude (mV)', 'Frequency (kHz)', 'Quality Factor', 'Phase (deg)',
                                              'R^2 Criterion'], cmap=plt.cm.viridis)

        if save_plots:
            plt_path = os.path.join(folder_path, basename + '_' + grp_name + 'Maps.png')
            fig_ms.savefig(plt_path, format='png', dpi=300)

    if show_plots:
        plt.show()

    plt.close('all')


def plot_loop_guess_fit(vdc, ds_proj_loops, ds_guess, ds_fit, title=''):
    """
    Plots the loop guess, fit, source projected loops for a single cycle

    Parameters
    ----------
    vdc - 1D float numpy array
        DC offset vector (unshifted)
    ds_proj_loops - 2D numpy array
        Projected loops arranged as [position, vdc]
    ds_guess - 1D compound numpy array
        Loop guesses arranged as [position]
    ds_fit - 1D compound numpy array
        Loop fits arranged as [position]
    title - (Optional) String / unicode
        Title for the figure

    Returns
    ----------
    fig - matplotlib.pyplot.figure object
        Figure handle
    axes - 2D array of matplotlib.pyplot.axis handles
        handles to axes in the 2d figure
    """
    shift_ind = int(-1 * len(vdc) / 4)
    vdc_shifted = np.roll(vdc, shift_ind)
    loops_shifted = np.roll(ds_proj_loops, shift_ind, axis=1)

    num_plots = np.min([5, int(np.sqrt(ds_proj_loops.shape[0]))])
    fig, axes = plt.subplots(nrows=num_plots, ncols=num_plots, figsize=(18, 18))
    positions = np.linspace(0, ds_proj_loops.shape[0] - 1, num_plots ** 2, dtype=np.int)
    for ax, pos in zip(axes.flat, positions):
        ax.plot(vdc_shifted, loops_shifted[pos, :], 'k', label='Raw')
        ax.plot(vdc_shifted, loop_fit_function(vdc_shifted, np.array(list(ds_guess[pos]))), 'g', label='guess')
        ax.plot(vdc_shifted, loop_fit_function(vdc_shifted, np.array(list(ds_fit[pos]))), 'r--', label='Fit')
        ax.set_xlabel('V_DC (V)')
        ax.set_ylabel('PR (a.u.)')
        ax.set_title('Position ' + str(pos))
    ax.legend()
    fig.suptitle(title)
    fig.tight_layout()

    return fig, axes


def jupyter_visualize_beps_sho(h5_sho_dset, step_chan, resp_func=None, resp_label='Response'):
    """
    Jupyer notebook ONLY function. Sets up an interactive visualizer for viewing SHO fitted BEPS data.
    Currently, this is limited to DC and AC spectroscopy datasets.

    Parameters
    ----------
    h5_sho_dset : h5py.Dataset
        dataset to be plotted
    step_chan : string / unicode
        Name of the channel that forms the primary spectroscopic axis (eg - DC offset)
    resp_func : function (optional)
        Function to apply to the spectroscopic data. Currently, DC spectroscopy uses A*cos(phi) and AC spectroscopy
        uses A
    resp_label : string / unicode (optional)
        Label for the response (y) axis.
    """
    guess_3d_data, success = reshape_to_Ndims(h5_sho_dset)

    h5_sho_spec_inds = getAuxData(h5_sho_dset, 'Spectroscopic_Indices')[0]
    h5_sho_spec_vals = getAuxData(h5_sho_dset, 'Spectroscopic_Values')[0]
    spec_nd, _ = reshape_to_Ndims(h5_sho_spec_inds, h5_spec=h5_sho_spec_inds)
    # sho_spec_sort = get_sort_order(h5_sho_spec_inds)
    sho_spec_dims = np.array(spec_nd.shape[1:])
    sho_spec_labels = get_attr(h5_sho_spec_inds, 'labels')

    h5_pos_inds = getAuxData(h5_sho_dset, auxDataName='Position_Indices')[-1]
    pos_nd, _ = reshape_to_Ndims(h5_pos_inds, h5_pos=h5_pos_inds)
    pos_dims = list(pos_nd.shape[:h5_pos_inds.shape[1]])
    pos_labels = get_attr(h5_pos_inds, 'labels')

    # reshape to X, Y, step, all others
    spec_step_dim_ind = np.where(sho_spec_labels == step_chan)[0][0]
    step_dim_ind = len(pos_dims) + spec_step_dim_ind

    # move the step dimension to be the first after all position dimensions
    rest_sho_dim_order = list(range(len(pos_dims), len(guess_3d_data.shape)))
    rest_sho_dim_order.remove(step_dim_ind)
    new_order = list(range(len(pos_dims))) + [step_dim_ind] + rest_sho_dim_order

    # Transpose the 3D dataset to this shape:
    sho_guess_Nd_1 = np.transpose(guess_3d_data, new_order)

    # Now move the step dimension to the front for the spec labels as well
    new_spec_order = list(range(len(sho_spec_labels)))
    new_spec_order.remove(spec_step_dim_ind)
    new_spec_order = [spec_step_dim_ind] + new_spec_order

    # new_spec_labels = sho_spec_labels[new_spec_order]
    new_spec_dims = np.array(sho_spec_dims)[new_spec_order]

    # Now collapse all additional dimensions
    final_guess_shape = pos_dims + [new_spec_dims[0]] + [-1]
    sho_dset_collapsed = np.reshape(sho_guess_Nd_1, final_guess_shape)

    # Get the bias matrix:
    bias_mat, _ = reshape_to_Ndims(h5_sho_spec_vals, h5_spec=h5_sho_spec_inds)
    bias_mat = np.transpose(bias_mat[spec_step_dim_ind], new_spec_order).reshape(sho_dset_collapsed.shape[len(pos_dims):])

    # This is just the visualizer:
    sho_quantity = 'Amplitude [V]'
    step_ind = 0
    row_ind = 1
    col_ind = 1

    def dc_spectroscopy_func(resp_vec):
        return resp_vec['Amplitude [V]'] * np.cos(resp_vec['Phase [rad]']) * 1E+3

    def ac_spectroscopy_func(resp_vec):
        return resp_vec['Amplitude [V]']

    if resp_func is None:
        if step_chan == 'DC_Offset':
            resp_func = dc_spectroscopy_func
            resp_label = 'A cos($\phi$) (a. u.)'
        else:
            resp_func = ac_spectroscopy_func
            resp_label = 'Amplitude (a. u.)'

    spatial_map = sho_dset_collapsed[:, :, step_ind, 0][sho_quantity]
    resp_vec = sho_dset_collapsed[row_ind, col_ind, :, :]
    resp_vec = resp_func(resp_vec)

    # bias_vec = bias_mat[:, loop_ind]

    fig = plt.figure(figsize=(12, 8))
    ax_bias = plt.subplot2grid((3, 2), (0, 0), colspan=1, rowspan=1)
    ax_map = plt.subplot2grid((3, 2), (1, 0), rowspan=2)
    ax_loop = plt.subplot2grid((3, 2), (0, 1), colspan=1, rowspan=3)

    ax_bias.plot(bias_mat[:, 0])
    ax_bias.set_xlabel('Bias Step')
    ax_bias.set_ylabel(step_chan.replace('_', ' ') + ' (V)')
    bias_slider = ax_bias.axvline(x=step_ind, color='r')

    img_map = ax_map.imshow(spatial_map, cmap=plt.cm.viridis, origin='lower',
                            interpolation='none')
    ax_map.set_xlabel('X')
    ax_map.set_ylabel('Y')
    main_vert_line = ax_map.axvline(x=row_ind, color='k')
    main_hor_line = ax_map.axhline(y=col_ind, color='k')

    ax_loop.axvline(x=0, color='gray', linestyle='--')
    ax_loop.axhline(y=0, color='gray', linestyle='--')
    line_handles = ax_loop.plot(bias_mat, resp_vec)
    ax_loop.set_xlabel(step_chan.replace('_', ' ') + ' (V)')
    ax_loop.set_ylabel(resp_label)
    fig.tight_layout()

    def update_sho_plots(sho_quantity, **kwargs):
        step_ind = kwargs['Bias Step']
        bias_slider.set_xdata((step_ind, step_ind))
        spatial_map = sho_dset_collapsed[:, :, step_ind, 0][sho_quantity]
        img_map.set_data(spatial_map)
        spat_mean = np.mean(spatial_map)
        spat_std = np.std(spatial_map)
        img_map.set_clim(vmin=spat_mean - 3 * spat_std, vmax=spat_mean + 3 * spat_std)

        row_ind = kwargs['Y']
        col_ind = kwargs['X']
        main_vert_line.set_xdata((kwargs['X'], kwargs['X']))
        main_hor_line.set_ydata((kwargs['Y'], kwargs['Y']))

        resp_vec = sho_dset_collapsed[row_ind, col_ind, :, :]
        resp_vec = resp_func(resp_vec)
        for line_handle, data in zip(line_handles, np.transpose(resp_vec)):
            line_handle.set_ydata(data)
        ax_loop.set_ylim([np.min(resp_vec), np.max(resp_vec)])
        display(fig)

    slider_dict = dict()
    for pos_dim_ind, dim_name in enumerate(pos_labels):
        slider_dict[dim_name] = (0, pos_dims[pos_dim_ind] - 1, 1)
    slider_dict['Bias Step'] = (0, bias_mat.shape[0] - 1, 1)

    widgets.interact(update_sho_plots, sho_quantity=list(sho_dset_collapsed.dtype.names[:-1]), **slider_dict)


def jupyter_visualize_be_spectrograms(h5_main):
    """
    Jupyer notebook ONLY function. Sets up a simple visualzier for visualizing raw BE data.
    Sliders for position indices can be used to visualize BE spectrograms (frequency, UDVS step).
    In the case of 2 spatial dimensions, a spatial map will be provided as well

    Parameters
    ----------
    h5_main : h5py.Dataset
        Raw dataset
    """
    # TODO add colormap options
    h5_pos_inds = getAuxData(h5_main, auxDataName='Position_Indices')[-1]
    pos_sort = get_sort_order(np.transpose(h5_pos_inds))
    pos_dims = get_dimensionality(np.transpose(h5_pos_inds), pos_sort)
    pos_labels = np.array(get_attr(h5_pos_inds, 'labels'))[pos_sort]

    h5_spec_vals = getAuxData(h5_main, auxDataName='Spectroscopic_Values')[-1]
    h5_spec_inds = getAuxData(h5_main, auxDataName='Spectroscopic_Indices')[-1]
    spec_sort = get_sort_order(h5_spec_inds)
    spec_dims = get_dimensionality(h5_spec_inds, spec_sort)
    spec_labels = np.array(get_attr(h5_spec_inds, 'labels'))[spec_sort]

    ifreq = int(np.argwhere(spec_labels == 'Frequency'))
    freqs_nd = reshape_to_Ndims(h5_spec_vals, h5_spec=h5_spec_inds)[0][ifreq].squeeze()
    freqs_2d = freqs_nd.reshape(freqs_nd.shape[0], -1) / 1000  # Convert to kHz
    try:
        num_udvs_steps = h5_main.parent.parent.attrs['num_udvs_steps']
    except KeyError:
        num_udvs_steps = h5_main.parent.parent.attrs['num_UDVS_steps']

    if len(pos_dims) == 2:
        spatial_map = np.abs(np.reshape(h5_main[:, 0], pos_dims))
        spectrogram = np.reshape(h5_main[0], (num_udvs_steps, -1))
        fig, axes = plt.subplots(ncols=3, figsize=(12, 4))
        # spatial_img = axes[0].imshow(np.abs(spatial_map), cmap=plt.cm.viridis)
        spatial_img = plot_map(axes[0], np.abs(spatial_map), origin='lower',
                               cmap=plt.cm.viridis)
        axes[0].set_xlabel('X')
        axes[0].set_ylabel('Y')
        main_vert_line = axes[0].axvline(x=int(0.5 * spatial_map.shape[1]), color='k')
        main_hor_line = axes[0].axhline(y=int(0.5 * spatial_map.shape[0]), color='k')
        amp_img = axes[1].imshow(np.abs(spectrogram), cmap=plt.cm.viridis,
                                 extent=[freqs_2d[0, 0], freqs_2d[-1, 0],
                                         spectrogram.shape[0], 0],
                                 interpolation='none')
        axes[1].set_title('Amplitude')
        axes[1].set_xlabel('Frequency (kHz)')
        axes[1].set_ylabel('BE step')
        phase_img = axes[2].imshow(np.angle(spectrogram), cmap=plt.cm.viridis,
                                   extent=[freqs_2d[0, 0], freqs_2d[-1, 0],
                                           spectrogram.shape[0], 0],
                                   interpolation='none')
        axes[2].set_title('Phase')
        axes[2].set_xlabel('Frequency (kHz)')
        for axis in axes[1:3]:
            axis.axis('tight')
            axis.set_ylim(0, spectrogram.shape[0])
        fig.tight_layout()

        def index_unpacker(**kwargs):
            spatial_map = np.abs(np.reshape(h5_main[:, kwargs['spectroscopic']], pos_dims))
            spatial_img.set_data(spatial_map)
            spat_mean = np.mean(spatial_map)
            spat_std = np.std(spatial_map)
            spatial_img.set_clim(vmin=spat_mean - 3 * spat_std, vmax=spat_mean + 3 * spat_std)

            spec_heading = ''
            for dim_ind, dim_name in enumerate(spec_labels):
                spec_heading += dim_name + ': ' + str(h5_spec_vals[dim_ind, kwargs['spectroscopic']]) + ', '
            axes[0].set_title(spec_heading[:-2])

            pos_dim_vals = list(range(len(pos_labels)))
            for pos_dim_ind, pos_dim_name in enumerate(pos_labels):
                pos_dim_vals[pos_dim_ind] = kwargs[pos_dim_name]

            main_vert_line.set_xdata((kwargs['X'], kwargs['X']))
            main_hor_line.set_ydata((kwargs['Y'], kwargs['Y']))

            pix_ind = pos_dim_vals[0]
            for pos_dim_ind in range(1, len(pos_labels)):
                pix_ind += pos_dim_vals[pos_dim_ind] * pos_dims[pos_dim_ind - 1]
            spectrogram = np.reshape(h5_main[pix_ind], (num_udvs_steps, -1))
            amp_img.set_data(np.abs(spectrogram))
            phase_img.set_data(np.angle(spectrogram))
            display(fig)

        pos_dict = dict()
        for pos_dim_ind, dim_name in enumerate(pos_labels):
            pos_dict[dim_name] = (0, pos_dims[pos_dim_ind] - 1, 1)
        pos_dict['spectroscopic'] = (0, h5_main.shape[1] - 1, 1)

        widgets.interact(index_unpacker, **pos_dict)
    else:
        def plot_spectrogram(data, freq_vals):
            fig, axes = plt.subplots(ncols=2, figsize=(9, 5), sharey=True)
            im_handles = list()
            im_handles.append(axes[0].imshow(np.abs(data), cmap=plt.cm.viridis,
                                             extent=[freqs_2d[0, 0], freqs_2d[-1, 0],
                                                     data.shape[0], 0],
                                             interpolation='none'))
            axes[0].set_title('Amplitude')
            axes[0].set_ylabel('BE step')
            im_handles.append(axes[1].imshow(np.angle(data), cmap=plt.cm.viridis,
                                             extent=[freqs_2d[0, 0], freqs_2d[-1, 0],
                                                     data.shape[0], 0],
                                             interpolation='none'))
            axes[1].set_title('Phase');
            axes[0].set_xlabel('Frequency index')
            axes[1].set_xlabel('Frequency index')
            for axis in axes:
                axis.axis('tight')
                axis.set_ylim(0, data.shape[0])
            fig.tight_layout()
            return fig, axes, im_handles

        fig, axes, im_handles = plot_spectrogram(np.reshape(h5_main[0], (num_udvs_steps, -1)), freqs_2d)

        def position_unpacker(**kwargs):
            pos_dim_vals = range(len(pos_labels))
            for pos_dim_ind, pos_dim_name in enumerate(pos_labels):
                pos_dim_vals[pos_dim_ind] = kwargs[pos_dim_name]
            pix_ind = pos_dim_vals[0]
            for pos_dim_ind in range(1, len(pos_labels)):
                pix_ind += pos_dim_vals[pos_dim_ind] * pos_dims[pos_dim_ind - 1]
            spectrogram = np.reshape(h5_main[pix_ind], (num_udvs_steps, -1))
            im_handles[0].set_data(np.abs(spectrogram))
            im_handles[1].set_data(np.angle(spectrogram))
            display(fig)

        pos_dict = dict()
        for pos_dim_ind, dim_name in enumerate(pos_labels):
            pos_dict[dim_name] = (0, pos_dims[pos_dim_ind] - 1, 1)

        widgets.interact(position_unpacker, **pos_dict)


def jupyter_visualize_beps_loops(h5_projected_loops, h5_loop_guess, h5_loop_fit, step_chan='DC_Offset'):
    """
    Interactive plotting of the BE Loops
    
    Parameters
    ----------
    h5_projected_loops : h5py.Dataset
        Dataset holding the loop projections
    h5_loop_guess : h5py.Dataset
        Dataset holding the loop guesses
    h5_loop_fit : h5py.Dataset
        Dataset holding the loop gits
    step_chan : str, optional
        The name of the Spectroscopic dimension to plot versus.  Needs testing.
        Default 'DC_Offset'

    Returns
    -------
    None
    
    """
    # Prepare some variables for plotting loops fits and guesses
    # Plot the Loop Guess and Fit Results
    proj_nd, _ = reshape_to_Ndims(h5_projected_loops)
    guess_nd, _ = reshape_to_Ndims(h5_loop_guess)
    fit_nd, _ = reshape_to_Ndims(h5_loop_fit)

    h5_projected_loops = h5_loop_guess.parent['Projected_Loops']
    h5_proj_spec_inds = getAuxData(h5_projected_loops,
                                                auxDataName='Spectroscopic_Indices')[-1]
    h5_proj_spec_vals = getAuxData(h5_projected_loops,
                                                auxDataName='Spectroscopic_Values')[-1]
    h5_pos_inds = getAuxData(h5_projected_loops,
                             auxDataName='Position_Indices')[-1]
    pos_nd, _ = reshape_to_Ndims(h5_pos_inds, h5_pos=h5_pos_inds)
    pos_dims = list(pos_nd.shape[:h5_pos_inds.shape[1]])
    pos_labels = get_attr(h5_pos_inds, 'labels')

    # reshape the vdc_vec into DC_step by Loop
    spec_nd, _ = reshape_to_Ndims(h5_proj_spec_vals, h5_spec=h5_proj_spec_inds)
    loop_spec_dims = np.array(spec_nd.shape[1:])
    loop_spec_labels = get_attr(h5_proj_spec_vals, 'labels')

    spec_step_dim_ind = np.where(loop_spec_labels == step_chan)[0][0]

    # # move the step dimension to be the first after all position dimensions
    rest_loop_dim_order = list(range(len(pos_dims), len(proj_nd.shape)))
    rest_loop_dim_order.pop(spec_step_dim_ind)
    new_order = list(range(len(pos_dims))) + [len(pos_dims)+spec_step_dim_ind] + rest_loop_dim_order

    new_spec_order = np.array(new_order[len(pos_dims):], dtype=np.uint32)-len(pos_dims)
    # new_spec_dims = loop_spec_dims[new_spec_order]
    # new_spec_labels = loop_spec_labels[new_spec_order]

    #Also reshape the projected loops to Positions-DC_Step-Loop
    final_loop_shape = pos_dims+[loop_spec_dims[spec_step_dim_ind]]+[-1]
    proj_nd_3 = np.reshape(proj_nd, final_loop_shape)

    # Do the same for the guess and fit datasets
    guess_3d = np.reshape(guess_nd, pos_dims+[-1])
    fit_3d = np.reshape(fit_nd, pos_dims + [-1])

    # Get the bias vector:
    bias_vec = np.reshape(spec_nd[spec_step_dim_ind], final_loop_shape[len(pos_dims):])

    # Shift the bias vector and the loops by a quarter cycle
    shift_ind = int(-1*bias_vec.shape[0]/4)
    bias_shifted = np.roll(bias_vec, shift_ind, axis=0)
    proj_nd_shifted = np.roll(proj_nd_3, shift_ind, axis=len(pos_dims))

    # This is just the visualizer:
    loop_field_names = fit_nd.dtype.names
    loop_field = loop_field_names[0]
    loop_ind = 0
    row_ind = 0
    col_ind = 0

    # Initial plot data
    spatial_map = fit_3d[:,:, loop_ind][loop_field]
    proj_data = proj_nd_shifted[row_ind, col_ind, :, loop_ind]
    bias_data = bias_shifted[:, loop_ind]
    guess_data = loop_fit_function(bias_data, np.array(list(guess_3d[row_ind, col_ind, loop_ind])))
    fit_data = loop_fit_function(bias_data, np.array(list(fit_3d[row_ind, col_ind, loop_ind])))

    fig = plt.figure(figsize=(12, 8))
    ax_map = plt.subplot2grid((1, 2), (0, 0), colspan=1, rowspan=1)
    ax_loop = plt.subplot2grid((1, 2), (0, 1), colspan=1, rowspan=1)

    im_map = ax_map.imshow(spatial_map, cmap=plt.cm.viridis,
                           origin='lower', interpolation='none')
    ax_map.set_xlabel('X')
    ax_map.set_ylabel('Y')
    main_vert_line = ax_map.axvline(x=row_ind, color='k')
    main_hor_line = ax_map.axhline(y=col_ind, color='k')

    ax_loop.plot(bias_data, proj_data, 'k', label='Projection')
    ax_loop.plot(bias_data, guess_data, 'g', label='Guess')
    ax_loop.plot(bias_data, fit_data, 'r--', label='Fit')
    line_handles = ax_loop.get_lines()

    ax_loop.tick_params(labelleft=False, labelright=True)
    ax_loop.yaxis.set_label_position('right')
    ax_loop.set_ylabel('PR (a.u.)')
    ax_loop.set_xlabel('V_DC (V)')
    ax_loop.set_title('Position ({},{})'.format(col_ind, row_ind))
    ax_loop.legend()
    fig.tight_layout()

    def update_loop_plots(loop_field, **kwargs):
        loop_ind = kwargs['Loop Number']
        spatial_map = fit_3d[:,:, loop_ind][loop_field]
        im_map.set_data(spatial_map)
        spat_mean = np.mean(spatial_map)
        spat_std = np.std(spatial_map)
        im_map.set_clim(vmin=spat_mean - 3 * spat_std, vmax=spat_mean + 3 * spat_std)

        row_ind = kwargs['Y']
        col_ind = kwargs['X']
        main_vert_line.set_xdata((kwargs['X'], kwargs['X']))
        main_hor_line.set_ydata((kwargs['Y'], kwargs['Y']))

        proj_data = proj_nd_shifted[row_ind, col_ind, :, loop_ind]
        bias_data = bias_shifted[:, loop_ind]
        guess_data = loop_fit_function(bias_data, np.array(list(guess_3d[row_ind, col_ind, loop_ind])))
        fit_data = loop_fit_function(bias_data, np.array(list(fit_3d[row_ind, col_ind, loop_ind])))
        for line_handle, data in zip(line_handles, [proj_data, guess_data, fit_data]):
            line_handle.set_ydata(data)
        ax_loop.set_ylim([np.min([proj_data, guess_data, fit_data]), np.max([proj_data, guess_data, fit_data])])
        ax_loop.set_title('Position ({},{})'.format(col_ind, row_ind))
        display(fig)

    slider_dict = dict()
    for pos_dim_ind, dim_name in enumerate(pos_labels):
        slider_dict[dim_name] = (0, pos_dims[pos_dim_ind] - 1, 1)
    slider_dict['Loop Number'] = (0, bias_vec.shape[1] - 1, 1)
    widgets.interact(update_loop_plots, loop_field=list(fit_nd.dtype.names), **slider_dict)



"""
Pycroscopy's analysis submodule

Submodules
----------

.. autosummary::
    :toctree: _autosummary

    be_loop_model
    be_sho_model
    fit_methods
    guess_methods
    model
    optimize

"""

from . import utils
from .utils import *
from . import be_sho_model
from .be_sho_model import BESHOmodel
from . import be_loop_model
from .be_loop_model import BELoopModel
from . import guess_methods
from .guess_methods import GuessMethods
from . import model
from .model import Model
from .optimize import Optimize
from . import fit_methods
from .fit_methods import *
__all__ = ['GuessMethods', 'Model', 'BESHOmodel', 'BELoopModel', 'utils', 'Optimize', 'fit_methods']
__all__ += utils.__all__



"""
Created on Feb 9, 2016

@author: Chris Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals

import os
from warnings import warn

import numpy as np
from skimage.measure import block_reduce
from skimage.util import crop

from .df_utils import dm4reader
from .df_utils.io_image import read_image, read_dm3, parse_dm4_parms
from .translator import Translator
from .utils import generate_dummy_main_parms, make_position_mat, get_spectral_slicing, \
    get_position_slicing, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, calc_chunks, link_as_main
from ..io_hdf5 import ioHDF5
from ..microdata import MicroDataGroup, MicroDataset


class OneViewTranslator(Translator):
    """
    Translate Pytchography data from a set of images to an HDF5 file
    """

    def __init__(self, *args, **kwargs):
        super(OneViewTranslator, self).__init__(*args, **kwargs)

        self.rebin = False
        self.bin_factor = 1
        self.hdf = None
        self.binning_func = self.__no_bin
        self.bin_func = None
        self.h5_main = None
        self.root_image_list = list()
        self.crop_method = 'percent'
        self.crop_ammount = None
        self.image_list_tag = None

    def translate(self, h5_path, image_path, bin_factor=None, bin_func=np.mean, start_image=0, scan_size_x=None,
                  scan_size_y=None, crop_ammount=None, crop_method='percent'):
        """
        Basic method that adds Ptychography data to existing hdf5 thisfile
        You must have already done the basic translation with BEodfTranslator

        Parameters
        ----------------
        h5_path : str
            Absolute path to where the HDF5 file should be located
        image_path : str
            Absolute path to folder holding the image files
        bin_factor : array_like of uint, optional
            Downsampling factor for each dimension.  Default is None.
        bin_func : callable, optional
            Function which will be called to calculate the return value
            of each block.  Function must implement an axis parameter,
            i.e. numpy.mean.  Ignored if bin_factor is None.  Default is
            numpy.mean.
        start_image : int, optional
            Integer denoting which image in the file path should be considered the starting
            point.  Default is 0, start with the first image on the list.
        scan_size_x : int, optional
            Number of Ronchigrams in the x direction.  Default is None, value will be determined
            from the number of images and `scan_size_y` if it is given.
        scan_size_y : int, optional
            Number of Ronchigrams in the y direction.  Default is None, value will be determined
            from the number of images and `scan_size_x` if it is given.
        crop_ammount : uint or list of uints, optional
            How much should be cropped from the original image.  Can be a single unsigned
            integer or a list of unsigned integers.  A single integer will crop the same
            ammount from all edges.  A list of two integers will crop the x-dimension by
            the first integer and the y-dimension by the second integer.  A list of 4
            integers will crop the image as [left, right, top, bottom].
        crop_method : str, optional
            Which cropping method should be used.  How much of the image is removed is
            determined by the value of `crop_ammount`.
            'percent' - A percentage of the image is removed.
            'absolute' - The specific number of pixel is removed.
        Returns
        ----------
        h5_main : h5py.Dataset
            HDF5 Dataset object that contains the flattened images

        """
        # Open the hdf5 file and delete any contents
        try:
            hdf = ioHDF5(h5_path)
            hdf.clear()
        except:
            raise

        self.hdf = hdf
        self.crop_method = crop_method
        self.crop_ammount = crop_ammount

        '''
        Get the list of all files with the .tif extension and
        the number of files in the list
        '''
        root_file_list, file_list = self._parse_file_path(image_path)

        size, image_parms = self._getimageparms(file_list[0])
        usize, vsize = size

        self.image_list_tag = image_parms.pop('Image_Tag', None)

        tmp, _ = read_image(file_list[0])
        if crop_ammount is not None:
            tmp = self.crop_ronc(tmp)
            usize, vsize = tmp.shape

        num_files = len(file_list)
        if scan_size_x is None and scan_size_y is None:
            scan_size_x = int(np.sqrt(num_files))
            scan_size_y = int(num_files/scan_size_x)
        elif scan_size_x is None:
            scan_size_x = int(num_files/scan_size_y)
        elif scan_size_y is None:
            scan_size_y = int(num_files/scan_size_x)

        '''
        Check if a bin_factor is given.  Set up binning objects if it is.
        '''
        if bin_factor is not None:
            self.rebin = True
            if isinstance(bin_factor, int):
                self.bin_factor = (bin_factor, bin_factor)
            elif len(bin_factor) == 2:
                self.bin_factor = tuple(bin_factor)
            else:
                raise ValueError('Input parameter `bin_factor` must be a length 2 array_like or an integer.\n' +
                                 '{} was given.'.format(bin_factor))
            usize = int(usize / self.bin_factor[0])
            vsize = int(vsize / self.bin_factor[1])
            self.binning_func = block_reduce
            self.bin_func = bin_func

        num_files = scan_size_x * scan_size_y

        h5_main, h5_mean_spec, h5_ronch = self._setupH5(usize, vsize, np.float32,
                                                        scan_size_x, scan_size_y,
                                                        image_parms)

        for root_file in root_file_list:
            print('Saving the root image located at {}.'.format(root_file))
            self._create_root_image(root_file)

        self._read_data(file_list[start_image:start_image + num_files],
                        h5_main, h5_mean_spec, h5_ronch, image_path)

        self.hdf.close()

        return

    def _create_root_image(self, image_path):
        """
        Create the Groups and Datasets for a single root image

        Parameters
        ----------
        image_path : str
            Path to the image file

        Returns
        -------
        None
        """
        image, image_parms = read_dm3(image_path)
        if image.ndim == 3:
            image = np.sum(image, axis=0)

        '''
        Create the Measurement and Channel Groups to hold the
        image Datasets
        '''
        root_grp = MicroDataGroup('/')

        meas_grp = MicroDataGroup('Measurement_')

        chan_grp = MicroDataGroup('Channel_')
        root_grp.addChildren([meas_grp])
        meas_grp.addChildren([chan_grp])

        '''
        Set the Measurement Group attributes
        '''
        meas_grp.attrs.update(image_parms)
        usize, vsize = image.shape
        meas_grp.attrs['image_size_u'] = usize
        meas_grp.attrs['image_size_v'] = vsize
        meas_grp.attrs['translator'] = 'OneView'
        meas_grp.attrs['num_pixels'] = image.size

        ds_rawimage = MicroDataset('Raw_Data', np.reshape(image, (-1, 1)))

        '''
        Build Spectroscopic and Position datasets for the image
        '''
        pos_mat = make_position_mat(image.shape)
        spec_mat = np.array([[0]], dtype=np.uint8)

        ds_spec_inds = MicroDataset('Spectroscopic_Indices', spec_mat)
        ds_spec_vals = MicroDataset('Spectroscopic_Values', spec_mat, dtype=np.float32)
        spec_lab = get_spectral_slicing(['Image'])
        ds_spec_inds.attrs['labels'] = spec_lab
        ds_spec_inds.attrs['units'] = ''
        ds_spec_vals.attrs['labels'] = spec_lab
        ds_spec_vals.attrs['units'] = ''

        ds_pos_inds = MicroDataset('Position_Indices', pos_mat)
        ds_pos_vals = MicroDataset('Position_Values', pos_mat, dtype=np.float32)

        pos_lab = get_position_slicing(['X', 'Y'])
        ds_pos_inds.attrs['labels'] = pos_lab
        ds_pos_inds.attrs['units'] = ['pixel', 'pixel']
        ds_pos_vals.attrs['labels'] = pos_lab
        ds_pos_vals.attrs['units'] = ['pixel', 'pixel']

        chan_grp.addChildren([ds_rawimage, ds_spec_inds, ds_spec_vals,
                              ds_pos_inds, ds_pos_vals])

        '''
        Write the data to file and get the handle for the image dataset
        '''
        image_refs = self.hdf.writeData(root_grp)

        h5_image = getH5DsetRefs(['Raw_Data'], image_refs)[0]

        '''
        Link references to raw
        '''
        aux_ds_names = ['Position_Indices', 'Position_Values', 'Spectroscopic_Indices', 'Spectroscopic_Values']
        link_as_main(h5_image, *getH5DsetRefs(aux_ds_names, image_refs))

        self.root_image_list.append(h5_image)

    def _read_data(self, file_list, h5_main, h5_mean_spec, h5_ronch, image_path):
        """
        Iterates over the images in `file_list`, reading each image and downsampling if
        reqeusted, and writes the flattened image to file.  Also builds the Mean_Ronchigram
        and the Spectroscopic_Mean datasets at the same time.

        Parameters
        ----------
        file_list : list of str
            List of all files in `image_path` that will be read
        h5_main : h5py.Dataset
            Dataset which will hold the Ronchigrams
        h5_mean_spec : h5py.Dataset
            Dataset which will hold the Spectroscopic Mean
        h5_ronch : h5py.Dataset
            Dataset which will hold the Mean Ronchigram
        image_path : str
            Absolute file path to the directory which hold the images

        Returns
        -------
        None
        """

        mean_ronch = np.zeros(h5_ronch.shape, dtype=np.float32)

        num_files = len(file_list)

        for ifile, thisfile in enumerate(file_list):

            selected = (ifile + 1) % int(round(num_files / 16)) == 0
            if selected:
                print('Processing file...{}% - reading: {}'.format(round(100 * ifile / num_files), thisfile))

            image, _ = read_image(os.path.join(image_path, thisfile), get_parms=False, header=self.image_list_tag)
            # image, _ = read_image(os.path.join(image_path, thisfile), get_parms=False)
            image = self.crop_ronc(image)
            image = self.binning_func(image, self.bin_factor, self.bin_func)
            image = image.flatten()
            h5_main[ifile, :] = image

            h5_mean_spec[ifile] = np.mean(image)

            mean_ronch += image

            self.hdf.flush()

        h5_ronch[:] = mean_ronch / num_files
        self.hdf.flush()

    def crop_ronc(self, ronc):
        """
        Crop the input Ronchigram by the specified ammount using the specified method.

        Parameters
        ----------
        ronc : numpy.array
            Input image to be cropped.

        Returns
        -------
        cropped_ronc : numpy.array
            Cropped image
        """

        if self.crop_ammount is None:
            return ronc

        crop_ammount = self.crop_ammount
        crop_method = self.crop_method

        if crop_method == 'percent':
            crop_ammount = np.round(np.atleast_2d(crop_ammount)/100.0*ronc.shape)
            crop_ammount = tuple([tuple(row) for row in crop_ammount.astype(np.uint32)])
        elif crop_method == 'absolute':
            if isinstance(crop_ammount, int):
                pass
            elif len(crop_ammount) == 2:
                crop_ammount = ((crop_ammount[0],), (crop_ammount[1],))
            elif len(crop_ammount) == 4:
                crop_ammount = ((crop_ammount[0], crop_ammount[1]), (crop_ammount[2], crop_ammount[3]))
            else:
                raise ValueError('The crop_ammount should be an integer or list of 2 or 4 integers.')
        else:
            raise ValueError('Allowed values of crop_method are percent and absolute.')

        cropped_ronc = crop(ronc, crop_ammount)

        if any([dim == 0 for dim in cropped_ronc.shape]):
            warn("Requested crop ammount is greater than the image size.  No cropping will be done.")
            return ronc

        return cropped_ronc

    def downSampRoncVec(self, ronch_vec, binning_factor):
        """
        Downsample the image by taking the mean over nearby values

        Parameters
        ----------
        ronch_vec : ndarray
            Image data
        binning_factor : int
            factor to reduce the size of the image by

        Returns
        -------
        ronc_mat3_mean : ndarray
            Flattened downsampled image
        """
        ccd_pix = int(np.sqrt(ronch_vec.size))
        ronc_mat = ronch_vec.reshape(ccd_pix, ccd_pix)
        ronc_mat2 = ronc_mat.reshape(ccd_pix, ccd_pix / binning_factor, binning_factor)
        ronc_mat2_mean = ronc_mat2.mean(2)  # take the mean along the 3rd dimension
        ronc_mat3 = ronc_mat2_mean.reshape(ccd_pix / binning_factor, binning_factor, -1)
        ronc_mat3_mean = ronc_mat3.mean(1)

        return ronc_mat3_mean.reshape(-1)

    @staticmethod
    def _parse_file_path(image_folder):
        """
        Returns a list of all files in the directory given by path

        Parameters
        ---------------
        path : string / unicode
            absolute path to directory containing files

        Returns
        ----------
        file_list : list of strings
            names of all files in directory located at path
        """
        file_list = list()
        root_file_list = list()
        allowed_image_types = ['.dm3', '.dm4', '.jpg', '.png', '.tif',
                               '.tiff', '.jpeg', '.bmp']
        for root, dirs, files in os.walk(image_folder):
            for thisfile in files:
                _, ext = os.path.splitext(thisfile)
                if ext not in allowed_image_types:
                    continue
                if root == image_folder:
                    root_file_list.append(os.path.join(image_folder, thisfile))
                else:
                    file_list.append(os.path.join(root, thisfile))
        return root_file_list, file_list

    @staticmethod
    def _getimageparms(image):
        """
        Returns the x and y size of the image in pixels

        Parameters
        ------------
        image : string / unicode
            absolute path to the dm4 file

        Returns
        -----------
        size : unsigned integer
            x and y dimenstions of image
        parms : dict
            Image parameters from the dm4 file
        """
        dm4_file = dm4reader.DM4File.open(image)
        tags = dm4_file.read_directory()
        parms = parse_dm4_parms(dm4_file, tags, '')

        u_size = parms['Root_ImageList_SubDir_000_ImageData_Dimensions_Tag_000']
        v_size = parms['Root_ImageList_SubDir_000_ImageData_Dimensions_Tag_001']
        size = u_size, v_size

        return size, parms

    def _setupH5(self, usize, vsize, data_type, scan_size_x, scan_size_y, image_parms):
        """
        Setup the HDF5 file in which to store the data including creating
        the Position and Spectroscopic datasets

        Parameters
        ----------
        usize : int
            Number of pixel columns in the images
        vsize : int
            Number of pixel rows in the images
        data_type : type
            Data type to save image as
        scan_size_x : int
            Number of images in the x dimension
        scan_size_y : int
            Number of images in the y dimension
        image_parms : dict
            Dictionary of parameters

        Returns
        -------
        h5_main : h5py.Dataset
            HDF5 Dataset that the images will be written into
        h5_mean_spec : h5py.Dataset
            HDF5 Dataset that the mean over all positions will be written
            into
        h5_ronch : h5py.Dataset
            HDF5 Dateset that the mean over all Spectroscopic steps will be
            written into
        """
        num_pixels = usize * vsize
        num_files = scan_size_x * scan_size_y

        root_parms = generate_dummy_main_parms()
        root_parms['data_type'] = 'PtychographyData'

        main_parms = {'num_images': num_files,
                      'image_size_u': usize,
                      'image_size_v': vsize,
                      'num_pixels': num_pixels,
                      'translator': 'Ptychography',
                      'scan_size_x': scan_size_x,
                      'scan_size_y': scan_size_y}
        main_parms.update(image_parms)

        # Create the hdf5 data Group
        root_grp = MicroDataGroup('/')
        root_grp.attrs = root_parms
        meas_grp = MicroDataGroup('Measurement_000')
        meas_grp.attrs = main_parms
        chan_grp = MicroDataGroup('Channel_000')
        # Get the Position and Spectroscopic Datasets
        ds_spec_ind, ds_spec_vals = build_ind_val_dsets((usize, vsize), is_spectral=True,
                                                        labels=['U', 'V'], units=['pixel', 'pixel'])
        ds_pos_ind, ds_pos_val = build_ind_val_dsets([scan_size_x, scan_size_y], is_spectral=False,
                                                     labels=['X', 'Y'], units=['pixel', 'pixel'])

        ds_chunking = calc_chunks([num_files, num_pixels],
                                  data_type(0).itemsize,
                                  unit_chunks=(1, num_pixels))

        # Allocate space for Main_Data and Pixel averaged Data
        ds_main_data = MicroDataset('Raw_Data', data=[], maxshape=(num_files, num_pixels),
                                    chunking=ds_chunking, dtype=data_type, compression='gzip')
        ds_mean_ronch_data = MicroDataset('Mean_Ronchigram',
                                          data=np.zeros(num_pixels, dtype=np.float32),
                                          dtype=np.float32)
        ds_mean_spec_data = MicroDataset('Spectroscopic_Mean',
                                         data=np.zeros(num_files, dtype=np.float32),
                                         dtype=np.float32)
        # Add datasets as children of Measurement_000 data group
        chan_grp.addChildren([ds_main_data, ds_spec_ind, ds_spec_vals, ds_pos_ind,
                              ds_pos_val, ds_mean_ronch_data, ds_mean_spec_data])
        meas_grp.addChildren([chan_grp])

        root_grp.addChildren([meas_grp])
        # print('Writing following tree to this file:')
        # root_grp.showTree()

        h5_refs = self.hdf.writeData(root_grp)
        h5_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
        h5_ronch = getH5DsetRefs(['Mean_Ronchigram'], h5_refs)[0]
        h5_mean_spec = getH5DsetRefs(['Spectroscopic_Mean'], h5_refs)[0]
        aux_ds_names = ['Position_Indices',
                        'Position_Values',
                        'Spectroscopic_Indices',
                        'Spectroscopic_Values']

        link_as_main(h5_main, *getH5DsetRefs(aux_ds_names, h5_refs))

        self.hdf.flush()

        return h5_main, h5_mean_spec, h5_ronch

    @staticmethod
    def __no_bin(image, *args, **kwargs):
        """
        Does absolutely nothing to the image.  Exists so that we can have
        a bin function to call whether we actually rebin the image or not.

        Parameters
        ----------
        image : ndarray
            Image
        args:
            Argument list
        kwargs:
            Keyword argument list

        Returns
        -------
        image : ndarray
            The input image
        """
        return image



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 31 17:03:29 2016

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import numpy as np

# TODO: Test and debug node and clusterTree classes for agglomerative clustering etc

class Node(object):
    """
    Basic unit of a tree - a node. Keeps track of its value, labels, parent, children, level in the tree etc.
    """
    def __init__(self, name, value=None, parent=None, dist=0, labels=None, children=[], compute_mean=False,
                 verbose=False):
        """
        Parameters
        ----------
        name : (Optional) unsigned int
            ID of this node
        value : (Optional) 1D numpy array
            Response corresponding to this Node.
        parent : (Optional) unsigned int or Node object
            Parent for this Node.
        dist : (Optional) float
            Distance between the children nodes
        labels : (Optional) list or 1D numpy array of unsigned integers
            Positions / instances in a main dataset within this cluster
        children : (Optional) list of Node objects
            Children for this node
        compute_mean : (Optional) Boolean
            Whether or not to compute the value attribute from the provided children
        """
        self.children = children
        self.parent = parent
        self.name = name
        self.value = value
        self.dist = dist
        self.level = 0

        # Assign this node as the parent for all its children
        for child in children:
            child.parent = self

        # If labels were not provided (tree node), get from children
        if labels is None:
            temp_labels = []
            for child in self.children:
                if verbose:
                    print('Child #{} had the following labels:'.format(child.name))
                    print(child.labels)
                temp_labels.append(np.array(child.labels))
            if verbose:
                print('Labels (unsorted) derived from children for node #{}:'.format(name))
                print(temp_labels)
            self.labels = np.hstack(temp_labels)
            self.labels.sort()
        else:
            if verbose:
                print('Labels for leaf node #{}:'.format(name))
                print(labels)
            self.labels = np.array(labels, dtype=np.uint32)

        # Compute the level for this node along with the number of children below it
        if len(self.children) > 0:
            self.num_nodes = 0
            for child in self.children:
                self.num_nodes += child.num_nodes
                self.level = max(self.level, child.level)
            self.level += 1  # because this node has to be one higher level than its highest children
        else:
            self.num_nodes = 1
            if verbose:
                print('Parent node:', str(name), 'has', str(self.num_nodes), 'children')
        if all([len(self.children) > 0, type(value) == type(None), compute_mean]):
            resp = []
            for child in children:
                if verbose:
                    print('       Child node', str(child.name), 'has', str(child.num_nodes), 'children')
                # primitive method of equal bias mean: resp.append(child.value)
                # weighted mean:
                resp.append(child.value * child.labels.size / self.labels.size)
                # self.value = np.mean(np.array(resp), axis=0)
            self.value = np.sum(np.array(resp), axis=0)

    def __str__(self):
        return '({}) --> {},{}'.format(self.name, str(self.children[0].name), str(self.children[1].name))


class ClusterTree(object):
    """
    Creates a tree representation from the provided linkage pairing. Useful for clustering
    """
    def __init__(self, linkage_pairing, labels, distances=None, centroids=None):
        """
        Parameters
        ----------
        linkage_pairing : 2D unsigned int numpy array or list
            Linkage pairing that describes a tree structure. The matrix should result in a single tree apex.
        labels : 1D unsigned int numpy array or list
            Labels assigned to each of the positions in the main dataset. Eg. Labels from clustering
        distances : (Optional) 1D numpy float array or list
            Distances between clusters
        centroids : (Optional) 2D numpy array
            Mean responses for each of the clusters. These will be propagated up
        """
        self.num_leaves = linkage_pairing.shape[0] + 1
        self.linkage = linkage_pairing
        self.centroids = centroids
        """ this list maintains pointers to the nodes pertaining to that cluster id for quick look-ups
        By default this lookup table just contains the number indices of these clusters.
        They will be replaced with node objects as and when the objects are created"""
        self.nodes = list()

        # now the labels is a giant list of labels assigned for each of the positions.
        self.labels = np.array(labels, dtype=np.uint32)
        """ the labels for the leaf nodes need to be calculated manually from the provided labels
        Populate the lowest level nodes / leaves first:"""
        for clust_id in range(self.num_leaves):
            which_pos = np.where(self.labels == clust_id)
            if centroids is not None:
                self.nodes.append(Node(clust_id, value=centroids[clust_id], labels=which_pos))
            else:
                self.nodes.append(Node(clust_id, labels=which_pos))

        for row in range(linkage_pairing.shape[0]):
            """print 'working on', linkage_pairing[row]
            we already have each of these children in our look-up table"""
            childs = []  # this is an empty list that will hold all the children corresponding to this node
            for col in range(linkage_pairing.shape[1]):
                """ look at each child in this row
                look up the node object corresponding to this label """
                childs.append(self.nodes[int(linkage_pairing[row, col])])
                # Now this row results in a new node. That is what we create here and assign the children to this node
            new_node = Node(row + self.num_leaves, children=childs, compute_mean=centroids is not None)
            # If distances are provided, add the distances attribute to this node.
            # This is the distance between the children
            if distances is not None:
                new_node.dist = distances[row]
                # add this node to the look-up table:
            self.nodes.append(new_node)

        self.tree = self.nodes[-1]

    def __str__(self):
        """
        Overrides the to string representation. Prints the names of the node and its children.
        Not very useful for large trees

        Returns
        --------
        String representation of the tree structure
        """
        return str(self.tree)



# -*- coding: utf-8 -*-
"""
Created on Thursday May 26 11:23:00 2016

@author:  Rama Vasudevan, Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals

from os import path, remove  # File Path formatting
from warnings import warn

import numpy as np # For array operations
from scipy.io.matlab import loadmat  # To load parameters stored in Matlab .mat file

from .df_utils.be_utils import trimUDVS, getSpectroscopicParmLabel, generatePlotGroups, createSpecVals, nf32
from .translator import Translator  # Because this class extends the abstract Translator class
from .utils import make_position_mat, get_position_slicing, generate_dummy_main_parms
from ..be_hdf_utils import maxReadPixels
from ..hdf_utils import getH5DsetRefs
from ..io_hdf5 import ioHDF5  # Now the translator is responsible for writing the data.
from ..microdata import MicroDataGroup, MicroDataset  # The building blocks for defining heirarchical storage in the H5 file


class BEodfRelaxationTranslator(Translator):
    """
    Translates old Relaxation data into the new H5 format. This is for the files generated from
    the old BEPSDAQ program utilizing two cards simultaneously.
    At present, this version of the translator only works for Out of field measurements
    It will not work for in-field. This should be fixed at a later date.
    
    """
  
    def translate(self, file_path, show_plots=True, save_plots=True, do_histogram=False):
        """
        Basic method that translates .dat data file(s) to a single .h5 file
        
        Inputs:
            file_path -- Absolute file path for one of the data files. 
            It is assumed that this file is of the OLD data format. 
            
        Outputs:
            Nothing
        """
        (folder_path, basename) = path.split(file_path)
        (basename, path_dict) = self._parse_file_path(file_path)
            
        h5_path = path.join(folder_path,basename+'.h5')

        isBEPS = True
        parm_dict = self.__getParmsFromOldMat(path_dict['old_mat_parms'])
    
        ignored_plt_grps = ['in-field'] #Here we assume that there is no in-field. 
                                        #If in-field data is captured then the translator would have to be modified.
        
        #Technically, we could do away with this if statement, as isBEPS is always true for this translation
        if isBEPS:
            parm_dict['data_type'] = 'BEPSData'
            
            std_expt = parm_dict['VS_mode'] != 'load user defined VS Wave from file'
            
            if not std_expt:
                warn('This translator does not handle user defined voltage spectroscopy')
                return
            
            spec_label = getSpectroscopicParmLabel(parm_dict['VS_mode']) 
            
        # Check file sizes:
        if 'read_real' in path_dict.keys():
            real_size = path.getsize(path_dict['read_real']);
            imag_size = path.getsize(path_dict['read_imag']);
        else:
            real_size = path.getsize(path_dict['write_real'])
            imag_size = path.getsize(path_dict['write_imag'])
            
        if real_size != imag_size:
            raise ValueError("Real and imaginary file sizes DON'T match!. Ending")

        add_pix = False         
        num_rows = int(parm_dict['grid_num_rows']);
        num_cols = int(parm_dict['grid_num_cols']);
        num_pix = num_rows*num_cols;
        tot_bins = real_size/(num_pix*4); #Finding bins by simple division of entire datasize
        
        #Check for case where only a single pixel is missing. 
        check_bins = real_size/((num_pix-1)*4) 
        
        if tot_bins % 1 and check_bins % 1: 
            warn('Aborting! Some parameter appears to have changed in-between')
            return;
        elif not tot_bins % 1:
#             Everything's ok
            pass
        elif not check_bins % 1:
            tot_bins = check_bins
            warn('Warning:  A pixel seems to be missing from the data.  File will be padded with zeros.') 
            add_pix = True 
        
        
        tot_bins = int(tot_bins)
        (bin_inds, bin_freqs, bin_FFT,ex_wfm,dc_amp_vec) = self.__readOldMatBEvecs(path_dict['old_mat_parms'])
        """Because this is the old data format and there is a discrepancy in the number of bins (they seem to be 2 less than the actual number),
        we need to re-calculate it based on the available data. This is done below."""
        
        band_width = parm_dict['BE_band_width_[Hz]']*(0.5 - parm_dict['BE_band_edge_trim'])
        st_f = parm_dict['BE_center_frequency_[Hz]'] - band_width
        en_f = parm_dict['BE_center_frequency_[Hz]'] + band_width   
        bin_freqs = np.linspace(st_f, en_f, len(bin_inds), dtype=np.float32)
            
        # Forcing standardized datatypes:
        bin_inds = np.int32(bin_inds)
        bin_freqs = np.float32(bin_freqs)
        bin_FFT = np.complex64(bin_FFT)
        ex_wfm = np.float32(ex_wfm)
            
        self.FFT_BE_wave = bin_FFT
        pos_mat = make_position_mat([num_cols, num_rows])
        pos_slices = get_position_slicing(['X', 'Y'], num_pix)
        
        ds_ex_wfm = MicroDataset('Excitation_Waveform', ex_wfm)     
        ds_pos_ind = MicroDataset('Position_Indices', pos_mat, dtype=np.uint32)
        ds_pos_ind.attrs['labels'] = pos_slices
        ds_pos_val = MicroDataset('Position_Values', np.float32(pos_mat))
        ds_pos_val.attrs['labels'] = pos_slices

        (UDVS_labs, UDVS_units, UDVS_mat) = self.__buildUDVSTable(parm_dict)
       
#       Remove the unused plot group columns before proceeding:
        (UDVS_mat, UDVS_labs, UDVS_units) = trimUDVS(UDVS_mat, UDVS_labs, UDVS_units, ignored_plt_grps)
       
        spec_inds = np.zeros(shape=(2,tot_bins), dtype=np.uint)
                  
#       Will assume that all excitation waveforms have same number of bins
        num_actual_udvs_steps = UDVS_mat.shape[0]/2#Here, the denominator is 2 because only out of field measruements. For IF + OF, should be 1
        bins_per_step = tot_bins/num_actual_udvs_steps
       
        #Some more checks
        if bins_per_step % 1:
            warn('Non integer number of bins per step!')
            return
        else:
            bins_per_step = int(bins_per_step)
            
        num_actual_udvs_steps = int(num_actual_udvs_steps)
           
        stind = 0           
        for step_index in xrange(UDVS_mat.shape[0]):  
            if UDVS_mat[step_index,2] < 1E-3: # invalid AC amplitude
                continue # skip
            spec_inds[0,stind:stind+bins_per_step] = np.arange(bins_per_step,dtype=np.uint32) # Bin step    
            spec_inds[1,stind:stind+bins_per_step] = step_index * np.ones(bins_per_step,dtype=np.uint32) # UDVS step
            stind += bins_per_step
        del stind,step_index
           
        # Some very basic information that can help the processing / analysis crew
        parm_dict['num_bins'] = tot_bins
        parm_dict['num_pix'] = num_pix
        parm_dict['num_udvs_steps'] = num_actual_udvs_steps
        
        udvs_slices = dict();
        for col_ind, col_name in enumerate(UDVS_labs):
            udvs_slices[col_name] = (slice(None),slice(col_ind,col_ind+1));          
        ds_UDVS = MicroDataset('UDVS', UDVS_mat)
        ds_UDVS.attrs['labels'] = udvs_slices
        ds_UDVS.attrs['units'] = UDVS_units
        
        ds_spec_mat = MicroDataset('Spectroscopic_Indices', spec_inds, dtype=np.uint32)
        ds_spec_mat.attrs['labels'] = {'UDVS_Step':(slice(1,2),slice(None)), 'Bin':(slice(0,1),slice(None))}         
        ds_bin_steps = MicroDataset('Bin_Step', np.arange(bins_per_step, dtype=np.uint32),dtype=np.uint32)
        
        # Need to add the Bin Waveform type - infer from UDVS        
        exec_bin_vec = self.signal_type*np.ones(len(bin_inds), dtype=np.int32)

        ds_bin_inds = MicroDataset('Bin_Indices', bin_inds, dtype=np.uint32)       
        ds_bin_freq = MicroDataset('Bin_Frequencies', bin_freqs)        
        ds_bin_FFT = MicroDataset('Bin_FFT', bin_FFT)
        ds_wfm_typ = MicroDataset('Bin_Wfm_Type', exec_bin_vec)
        
        # Create Spectroscopic Values and Spectroscopic Values Labels datasets
        spec_vals, spec_vals_labs, spec_vals_units =  createSpecVals(UDVS_mat, spec_inds, bin_freqs, parm_dict, UDVS_labs, UDVS_units)
        
        spec_vals_slices = dict()
        for row_ind, row_name in enumerate(spec_vals_labs):
            spec_vals_slices[row_name]=(slice(row_ind,row_ind+1),slice(None))
        ds_spec_vals_mat = MicroDataset('Spectroscopic_Values',np.array(spec_vals,dtype=np.float32))
        ds_spec_vals_mat.attrs['labels'] = spec_vals_slices
        ds_spec_vals_mat.attrs['units'] = spec_vals_units
        
        # Noise floor should be of shape: (udvs_steps x 3 x positions)
        ds_noise_floor = MicroDataset('Noise_Floor', np.zeros(shape=(num_pix, num_actual_udvs_steps), dtype=nf32),
                                      chunking=(1, num_actual_udvs_steps))
        # noise_labs = ['super_band','inter_bin_band','sub_band']
        # noise_slices = dict()
        # for col_ind, col_name in enumerate(noise_labs):
        #     noise_slices[col_name] = (slice(None),slice(col_ind,col_ind+1), slice(None))
        # ds_noise_floor.attrs['labels'] = noise_slices
        
        """ 
        ONLY ALLOCATING SPACE FOR MAIN DATA HERE!
        Chunk by each UDVS step - this makes it easy / quick to:
            1. read data for a single UDVS step from all pixels
            2. read an entire / multiple pixels at a time
        The only problem is that a typical UDVS step containing 50 steps occupies only 400 bytes.
        This is smaller than the recommended chunk sizes of 10,000 - 999,999 bytes
        meaning that the metadata would be very substantial.
        This assumption is fine since we almost do not handle any user defined cases
        """       

        """
        New Method for chunking the Main_Data dataset.  Chunking is now done in N-by-N squares of UDVS steps by pixels.
        N is determined dinamically based on the dimensions of the dataset.  Currently it is set such that individual
        chunks are less than 10kB in size.
        
        Chris Smith -- csmith55@utk.edu
        """
        pixel_chunking = maxReadPixels(10240, num_pix*num_actual_udvs_steps, bins_per_step, np.dtype('complex64').itemsize)
        chunking = np.floor(np.sqrt(pixel_chunking))
        chunking = max(1, chunking)
        chunking = min(num_actual_udvs_steps, num_pix, chunking)
        ds_main_data = MicroDataset('Raw_Data', data=[], maxshape=(num_pix,tot_bins), dtype=np.complex64, chunking=(chunking,chunking*bins_per_step), compression='gzip')
        
        chan_grp = MicroDataGroup('Channel_')
        chan_grp.attrs['Channel_Input'] = parm_dict['IO_Analog_Input_1']
        chan_grp.addChildren([ds_main_data, ds_noise_floor])
        chan_grp.addChildren([ds_ex_wfm, ds_pos_ind, ds_pos_val, ds_spec_mat, ds_UDVS,
                              ds_bin_steps, ds_bin_inds, ds_bin_freq, ds_bin_FFT,ds_wfm_typ,ds_spec_vals_mat])
        
        # technically should change the date, etc.
        meas_grp = MicroDataGroup('Measurement_')
        meas_grp.attrs = parm_dict
        meas_grp.addChildren([chan_grp])
        
        spm_data = MicroDataGroup('')
        global_parms = generate_dummy_main_parms()
        global_parms['grid_size_x'] = parm_dict['grid_num_cols'];
        global_parms['grid_size_y'] = parm_dict['grid_num_rows'];
        global_parms['experiment_date'] = parm_dict['File_date_and_time']
       
       # assuming that the experiment was completed:
        global_parms['current_position_x'] = parm_dict['grid_num_cols']-1;
        global_parms['current_position_y'] = parm_dict['grid_num_rows']-1;
        global_parms['data_type'] = parm_dict['data_type'] #self.__class__.__name__
        global_parms['translator'] = 'ODF'
            
        spm_data.attrs = global_parms
        spm_data.addChildren([meas_grp])

        if path.exists(h5_path):
            remove(h5_path)
        
        # Write everything except for the main data.
        self.hdf = ioHDF5(h5_path)
        #self.hdf.clear() #Doesn't seem to work
        
        h5_refs = self.hdf.writeData(spm_data)
                    
        self.ds_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
            
        #Now doing linkrefs:
        aux_ds_names = ['Excitation_Waveform', 'Position_Indices','Position_Values',
                     'Spectroscopic_Indices','UDVS','Bin_Step', 'Bin_Indices', 
                     'Bin_Frequencies','Bin_FFT','Bin_Wfm_Type','Noise_Floor', 'Spectroscopic_Values']
        self.hdf.linkRefs(self.ds_main, getH5DsetRefs(aux_ds_names, h5_refs))
        
        self.mean_resp = np.zeros(shape=(self.ds_main.shape[1]), dtype=np.complex64)
        self.max_resp = np.zeros(shape=(self.ds_main.shape[0]), dtype=np.float32)
        self.min_resp = np.zeros(shape=(self.ds_main.shape[0]), dtype=np.float32)
        
        # Now read the raw data files:
        self._read_data(path_dict['read_real'], path_dict['read_imag'], parm_dict)
        self.hdf.flush()
        
        generatePlotGroups(self.ds_main, self.hdf, self.mean_resp, folder_path, basename,
                           self.max_resp, self.min_resp, max_mem_mb=self.max_ram,
                           spec_label = spec_label, show_plots = show_plots, save_plots=save_plots,
                           do_histogram=do_histogram, ignore_plot_groups=ignored_plt_grps) #We ignored in-field plot group.
        
        
        self.hdf.close()
        
        return h5_path
            
    def _read_data(self, real_path, imag_path, parm_dict):
        """
        Reads the imaginary and real data files one pixel at a time andwrites to the H5 dataset.
        
        Inputs:
            real_path -- file path of the .dat file containing the real component of the data
            imag_path -- file path of the .dat file containing the imaginary component of the data
            parm_dict--dictionary of parameters for the experiment            
            
        Outputs: None
        """
        print('---- reading data one pixel at a time----------')                
        
        num_pix = int(parm_dict['grid_num_rows'])*int(parm_dict['grid_num_cols'])
        #print 'Number of rows is: ', parm_dict['grid_num_rows']
        #print 'Number of cols is: ', parm_dict['grid_num_cols']
        #print 'Rows * cols is:', int(parm_dict['grid_num_rows'])*int(parm_dict['grid_num_cols'])
        
        if path.getsize(real_path) != path.getsize(imag_path):
            print('Sizes of real and imaginary files NOT matching!!!!')
        if 1.0*path.getsize(real_path) % num_pix != 0:
            print('Incomplete dataset!!!')
      
        bytes_per_pix = path.getsize(real_path)/num_pix        
        f_real = open(real_path, "rb")
        f_imag = open(imag_path, "rb")
        
        for pix_ind in xrange(num_pix):

            print('Reading pixel #{}, file position {}'.format(pix_ind,hex(pix_ind*bytes_per_pix)))
            #f_real.seek(pix_ind*bytes_per_pix,0) #This is unnecessary step and fails for large datasets for an unknown reason.
            #f_imag.seek(pix_ind*bytes_per_pix,0)
            pix_vec = np.fromstring( f_real.read(int(bytes_per_pix)), dtype='f') + 1j*np.fromstring(f_imag.read(int(bytes_per_pix)), dtype='f') 
            #verbose=0
            # Make chronologically correct
         #   print("Reshape values:", parm_dict['BE_bins_per_read'],parm_dict['VS_steps_per_full_cycle'],parm_dict['BE_repeats'])
            pix_mat = np.reshape(pix_vec, (parm_dict['BE_bins_per_read'],parm_dict['VS_steps_per_full_cycle'],parm_dict['BE_repeats']))
            pix_mat_temp = np.transpose(pix_mat, (1,2,0))
            pix_vec2 = np.reshape(pix_mat_temp, -1)
            
            #Calculate the mean, min, max
            self.max_resp[pix_ind] = np.max(np.abs(pix_vec2))
            self.min_resp[pix_ind] = np.min(np.abs(pix_vec2))
            self.mean_resp = (1/(pix_ind+1))*(pix_vec2 + pix_ind*self.mean_resp)

            # Write to file now
            self.ds_main[pix_ind,:] = np.complex64(pix_vec2)
            self.hdf.flush()
            
        f_real.close()
        f_imag.close()
        
        print('Finished writing data to .h5')

        
        
    def __readOldMatBEvecs(self,file_path):
        """
    Returns information about the excitation BE waveform present in the .mat file
    
    Inputs:
        filepath -- Absolute filepath of the .mat parameter file
    
    Outputs:
        Tuple -- (bin_inds, bin_w, bin_FFT, BE_wave, dc_amp_vec_full)\n
        bin_inds -- Bin indices\n
        bin_w -- Excitation bin Frequencies\n
        bin_FFT -- FFT of the BE waveform for the excited bins\n
        BE_wave -- Band Excitation waveform\n
        dc_amp_vec_full -- spectroscopic waveform. 
        This information will be necessary for fixing the UDVS for AC modulation for example
        """
    
        matread = loadmat(file_path, squeeze_me=True)    
        BE_wave = matread['BE_wave_1']
        bin_inds = matread['bin_ind_s'] -1 # Python base 0. note also _s, for this case
        bin_w = matread['bin_w']
        dc_amp_vec_full = matread['dc_amp_vec_full']
        FFT_full = np.fft.fftshift(np.fft.fft(BE_wave))
        bin_FFT = np.conjugate(FFT_full[bin_inds]);
        return (bin_inds, bin_w, bin_FFT, BE_wave, dc_amp_vec_full)
        
    def _parse_file_path(self, data_filepath):
        """
        Returns the basename and a dictionary containing the absolute file paths for the
        real and imaginary data files, text and mat parameter files in a dictionary
        
        Inputs:
            data_filepath: Absolute path of the real / imaginary data file (.dat)
        Outputs:
            Tuple (basename, path_dict)
        """
        (folder_path, basename) = path.split(data_filepath)
        (super_folder, basename) = path.split(folder_path) 

        if basename.endswith('_c'):
            # Old old data format where the folder ended with a _d for some reason
            base_name = basename[:-2]
            
        """
        A single pair of real and imaginary files are / were generated for:
            BE-Line and BEPS (compiled version only generated out-of-field or 'read')
        Two pairs of real and imaginary files were generated for later BEPS datasets
            These have 'read' and 'write' prefixes to denote out or in field respectively
        """
        path_dict = dict()
              
        real_path = path.join(folder_path, base_name + '_sub_real.dat')
        imag_path = path.join(folder_path, base_name + '_sub_imag.dat')
        
        path_dict['read_real'] = real_path
        path_dict['read_imag'] = imag_path
        path_dict['old_mat_parms'] = data_filepath   
        
        return (basename, path_dict)
        
        
    def __getParmsFromOldMat(self,file_path):
        """
        Formats parameters found in the old parameters .mat file into a dictionary
        as though the dataset had a parms.txt describing it
        
        Inputs:
            file_path -- absolute filepath of the .mat file containing the parameters
            
        Outputs -- dictionary containing parameters
        """
        parm_dict = dict()
        matread = loadmat(file_path, squeeze_me=True)
        
        parm_dict['IO_rate'] = str(int(matread['AO_rate']/1E+6)) + ' MHz'
        
        position_vec = matread['position_vec']
        parm_dict['grid_current_row'] = position_vec[0]
        parm_dict['grid_current_col'] = position_vec[1]
        parm_dict['grid_num_rows'] = int(position_vec[2])
        parm_dict['grid_num_cols'] = int(position_vec[3])
        
        if position_vec[0] != position_vec[1] or position_vec[2] != position_vec[3]:
            warn('WARNING: Incomplete dataset. Translation not guaranteed!')
            parm_dict['grid_num_rows'] = int(position_vec[0]) # set to number of present cols and rows
            parm_dict['grid_num_cols'] = int(position_vec[1])
    
        BE_parm_vec_1 = matread['BE_parm_vec_1']
        # Not required for translation but necessary to have
        if BE_parm_vec_1[0] == 3:
            parm_dict['BE_phase_content'] = 'chirp-sinc hybrid'
        else:
            parm_dict['BE_phase_content'] = 'Unknown'
        parm_dict['BE_center_frequency_[Hz]'] = BE_parm_vec_1[1]
        parm_dict['BE_band_width_[Hz]'] = BE_parm_vec_1[2]
        parm_dict['BE_amplitude_[V]'] = BE_parm_vec_1[3]   
        parm_dict['BE_band_edge_trim'] = -1 * BE_parm_vec_1[6] # 150 most likely
        parm_dict['BE_phase_variation'] = BE_parm_vec_1[5] # 0.01 most likely        
        parm_dict['BE_repeats'] = 2**int(BE_parm_vec_1[8])
        parm_dict['File_date_and_time']=0 #For now ignoring.
        parm_dict['BE_bins_per_read'] = matread['bins_per_band_s']
        assembly_parm_vec = matread['assembly_parm_vec']
        
        
        
        if assembly_parm_vec[2] == 0:
            parm_dict['VS_measure_in_field_loops'] = 'out-of-field'
        elif assembly_parm_vec[2] == 1:
            parm_dict['VS_measure_in_field_loops'] = 'in and out-of-field'
        else:
            parm_dict['VS_measure_in_field_loops'] = 'in-field'
        
        parm_dict['IO_Analog_Input_1'] = '+/- 10V, FFT'
        if assembly_parm_vec[3] == 0:
            parm_dict['IO_Analog_Input_2'] = 'off'
        else:
            parm_dict['IO_Analog_Input_2'] = '+/- 10V, FFT'
            
        #num_driving_bands = assembly_parm_vec[0] # 0 = 1, 1 = 2 bands
        #band_combination_order = assembly_parm_vec[1] # 0 parallel 1 series
        
        VS_parms = matread['SS_parm_vec']
        dc_amp_vec_full = matread['dc_amp_vec_full']
           
        VS_start_V = VS_parms[4] 
        VS_start_loop_amp = VS_parms[5] 
        VS_final_loop_amp = VS_parms[6] 
        #VS_read_write_ratio = VS_parms[8] #1 <- SS_read_write_ratio
        
        parm_dict['VS_set_pulse_amplitude_[V]'] = VS_parms[9] #0 <- SS_set_pulse_amp 
        parm_dict['VS_read_voltage_[V]'] = VS_parms[3] 
        parm_dict['VS_steps_per_full_cycle'] = len(dc_amp_vec_full)#VS_parms[7]
        parm_dict['VS_cycle_fraction'] = 'full'
        parm_dict['VS_cycle_phase_shift'] = 0 
        parm_dict['VS_number_of_cycles'] = VS_parms[2]
        parm_dict['FORC_num_of_FORC_cycles']=1
        parm_dict['FORC_V_high1_[V]']=0
        parm_dict['FORC_V_high2_[V]']=0
        parm_dict['FORC_V_low1_[V]']=0
        parm_dict['FORC_V_low2_[V]']=0
        
        if VS_parms[0] == 0:
            parm_dict['VS_mode'] = 'DC modulation mode'
            parm_dict['VS_amplitude_[V]'] = 0.5*(max(dc_amp_vec_full) - min(dc_amp_vec_full))# VS_parms[1] # SS_max_offset_amplitude
            parm_dict['VS_offset_[V]'] = max(dc_amp_vec_full) + min(dc_amp_vec_full)     
        elif VS_parms[0] == 1:
            # FORC
            parm_dict['VS_mode'] = 'DC modulation mode'
            parm_dict['VS_amplitude_[V]'] = 1 # VS_parms[1] # SS_max_offset_amplitude
            parm_dict['VS_offset_[V]'] = 0
            parm_dict['VS_number_of_cycles'] = 1                             
            parm_dict['FORC_num_of_FORC_cycles']=VS_parms[2]     
            parm_dict['FORC_V_high1_[V]']=VS_start_V
            parm_dict['FORC_V_high2_[V]']=VS_start_V
            parm_dict['FORC_V_low1_[V]']=VS_start_V - VS_start_loop_amp
            parm_dict['FORC_V_low2_[V]']=VS_start_V - VS_final_loop_amp
        elif VS_parms[0] == 2:
            # AC mode 
            parm_dict['VS_mode'] = 'AC modulation mode with time reversal'
            parm_dict['VS_amplitude_[V]'] = 0.5*(VS_final_loop_amp)
            parm_dict['VS_offset_[V]'] = 0 # this is not correct. Fix manually when it comes to UDVS generation?                     
        else:
            parm_dict['VS_mode'] = 'Custom'
    
        return parm_dict
               
    def __buildUDVSTable(self,parm_dict):
        """
        Generates the UDVS table using the parameters
        
        Inputs:
            parm_dict -- Dictionary of parameters present in the text files
        
        Outputs:
            tuple (labels, table)
            labels -- List of strings - Labels for columns in the UDVS table
            table -- UDVS data table
        """
    
        def translateVal(target, strvals, numvals):
            """
            Internal function - Interprets the provided value using the provided lookup table
            """
        
            if len(strvals) is not len(numvals):
                return None    
            for strval, fltval in zip(strvals,numvals):
                if target == strval:
                    return fltval
            return None # not found in list
            
        #% Extract values from parm text file    
        BE_signal_type =1# translateVal(parm_dict['BE_phase_content'], ['chirp-sinc hybrid','1/2 harmonic excitation','1/3 harmonic excitation','pure sine'],[1,2,3,4])
        # This is necessary when normalzing the AI by the AO
        self.harmonic = BE_signal_type
        self.signal_type = BE_signal_type
       
        BE_amp = parm_dict['BE_amplitude_[V]']
        
        VS_amp = parm_dict['VS_amplitude_[V]']
        VS_offset = parm_dict['VS_offset_[V]']
        VS_steps = parm_dict['VS_steps_per_full_cycle']
        VS_cycles = parm_dict['VS_number_of_cycles']
        VS_fraction = translateVal(parm_dict['VS_cycle_fraction'], ['full','1/2','1/4','3/4'], [1.,0.5,0.25,0.75])
        VS_shift = parm_dict['VS_cycle_phase_shift']
        if VS_shift is not 0:
            VS_shift = translateVal(VS_shift,['1/4','1/2','3/4'],[0.25,0.5,0.75])
        VS_in_out_cond = translateVal(parm_dict['VS_measure_in_field_loops'], ['out-of-field','in-field','in and out-of-field'],[0,1,2])
        VS_ACDC_cond = translateVal(parm_dict['VS_mode'], ['DC modulation mode','AC modulation mode with time reversal','load user defined VS Wave from file','current mode'],[0,2,3,4])
        self.expt_type = VS_ACDC_cond
        FORC_cycles = parm_dict['FORC_num_of_FORC_cycles']
        FORC_A1 = parm_dict['FORC_V_high1_[V]']
        FORC_A2 = parm_dict['FORC_V_high2_[V]']
        FORC_B1 = parm_dict['FORC_V_low1_[V]']
        FORC_B2 = parm_dict['FORC_V_low2_[V]']
            
        #% build vector of voltage spectroscopy values
        
        if VS_ACDC_cond == 0 or VS_ACDC_cond == 4: #DC voltage spectroscopy or current mode 
            VS_amp_vec_1 = np.arange(0,1+1/(VS_steps/4), 1/(VS_steps/4) )
            VS_amp_vec_2 = np.flipud(VS_amp_vec_1[:-1])
            VS_amp_vec_3 = -VS_amp_vec_1[1:]
            VS_amp_vec_4 =  VS_amp_vec_1[1:-1]-1
            VS_amp_vec = VS_amp*(np.hstack((VS_amp_vec_1, VS_amp_vec_2,  VS_amp_vec_3, VS_amp_vec_4)))
            VS_amp_vec = np.roll(VS_amp_vec,int(np.floor(VS_steps/VS_fraction*VS_shift)))#apply phase shift to VS wave
            VS_amp_vec = VS_amp_vec[:int(np.floor(VS_steps*VS_fraction))] # cut VS waveform
            VS_amp_vec = np.tile(VS_amp_vec,VS_cycles) # repeat VS waveform
            VS_amp_vec = VS_amp_vec+VS_offset
                       
        if FORC_cycles > 1:
            VS_amp_vec = VS_amp_vec/np.max(np.abs(VS_amp_vec))
            FORC_cycle_vec = np.arange(0, FORC_cycles+1, FORC_cycles/(FORC_cycles-1))
            FORC_A_vec = FORC_cycle_vec*(FORC_A2-FORC_A1)/FORC_cycles + FORC_A1
            FORC_B_vec = FORC_cycle_vec*(FORC_B2-FORC_B1)/FORC_cycles + FORC_B1
            FORC_amp_vec = (FORC_A_vec-FORC_B_vec)/2
            FORC_off_vec = (FORC_A_vec+FORC_B_vec)/2
            
            VS_amp_mat = np.tile(VS_amp_vec,[FORC_cycles,1])
            FORC_amp_mat = np.tile(FORC_amp_vec,[len(VS_amp_vec),1]).transpose()
            FORC_off_mat = np.tile(FORC_off_vec,[len(VS_amp_vec),1]).transpose()
            VS_amp_mat = VS_amp_mat*FORC_amp_mat + FORC_off_mat
            VS_amp_vec = VS_amp_mat.reshape(int(FORC_cycles*VS_cycles*VS_fraction*VS_steps))
        
        BE_repeats = parm_dict['BE_repeats']
        total_steps = len(VS_amp_vec)*BE_repeats #Needed for relaxation datasets
                
        #% Build UDVS table:        
        if VS_ACDC_cond is 0 or VS_ACDC_cond is 4: #relaxation measurements
            
            num_VS_steps = total_steps*2 #To account for IF and OOF
                        
            UD_VS_table_label = ['step_num','dc_offset','ac_amp','wave_type','wave_mod','in-field','out-of-field']
            UD_VS_table = np.zeros(shape=(num_VS_steps,7), dtype=np.float32)
            UD_VS_table_unit = ['', 'V', 'A', '', '', 'V', 'V']
            
            UD_VS_table[:,0] = np.arange(0, num_VS_steps) # Python base 0
            for step_num in np.arange(0,VS_steps): 
                step_values = (np.arange(int(step_num)*int(BE_repeats)*2, (int(step_num)+1)*int(BE_repeats)*2))
                UD_VS_table[step_values,1] = VS_amp_vec[step_num]                            
                
            BE_IF_switch = np.abs(np.imag(np.exp(1j*np.pi/2*np.arange(1,num_VS_steps+1))))
            BE_OF_switch = np.abs(np.real(np.exp(1j*np.pi/2*np.arange(1,num_VS_steps+1))))
            
            if VS_in_out_cond is 0: # out of field only
                UD_VS_table[:,2] = BE_amp * BE_OF_switch
            elif VS_in_out_cond is 1: # in field only
                UD_VS_table[:,2] = BE_amp * BE_IF_switch
            elif VS_in_out_cond is 2: # both in and out of field
                UD_VS_table[:,2] = BE_amp * np.ones(num_VS_steps)
            
            UD_VS_table[:,3] = np.ones(num_VS_steps) # wave type
            UD_VS_table[:,4] = np.ones(num_VS_steps) * BE_signal_type # wave mod
            
            UD_VS_table[:,5] = float('NaN')*np.ones(num_VS_steps)
            UD_VS_table[:,6] = float('NaN')*np.ones(num_VS_steps)
                            
            UD_VS_table[BE_IF_switch==1,5] = UD_VS_table[BE_IF_switch==1,1]            
            UD_VS_table[BE_OF_switch==1,6] = UD_VS_table[BE_IF_switch==1,1]
            
        return (UD_VS_table_label, UD_VS_table_unit, UD_VS_table)
        


"""
Created on Mar 1, 2016

@author: Chris Smith -- cmsith55@utk.edu
"""

from __future__ import division, print_function, absolute_import
import numpy as np
import sys

if sys.version_info.major == 3 and sys.version_info.minor == 6:
    disable_histogram = True
else:
    disable_histogram = False
    from numpy_groupies import aggregate_np

def buildHistogram(x_hist, data_mat, N_x_bins, N_y_bins, weighting_vec=1, min_resp=None, max_resp=None, func=None,
                   debug=False, *args, **kwargs):
    """
    Creates histogram for a single block of pixels

    Parameters
    ----------
    x_hist : 1D numpy array
        bins for x-axis of 2d histogram
    data_mat : numpy array
        data to be binned for y-axis of 2d histogram
    weighting_vec : 1D numpy array or float
        weights. If setting all to one value, can be a scalar
    N_x_bins : integer
        number of bins in the x-direction
    N_y_bins : integer
        number of bins in the y-direction
    min_resp : float
        minimum value for y binning
    max_resp : float
        maximum value for y binning
    func : function
        function to be used to bin data_vec.  All functions should take as input data_vec.
        Arguments should be passed properly to func.  This has not been heavily tested.

    Returns
    -------
    pixel_hist : 2D numpy array
        contains the histogram of the input data

    Apply func to input data, convert to 1D array, and normalize
    """
    if func is not None:
        y_hist = func(data_mat, *args, **kwargs)
    else:
        y_hist = data_mat

    '''
    Get the min_resp and max_resp from y_hist if they are none
    '''
    if min_resp is None:
        min_resp = np.min(y_hist)
    if max_resp is None:
        max_resp = np.max(y_hist)
    if debug: print('min_resp', min_resp, 'max_resp', max_resp)

    y_hist = __scale_and_discretize(y_hist, N_y_bins, max_resp, min_resp, debug)

    '''
    Combine x_hist and y_hist into one matrix
    '''
    if debug:
        print(np.shape(x_hist))
        print(np.shape(y_hist))

    try:
        group_idx = np.zeros((2, x_hist.size), dtype=np.int32)
        group_idx[0, :] = x_hist
        group_idx[1, :] = y_hist
    except:
        raise

    '''
    Aggregate matrix for histogram of current chunk
    '''
    if debug:
        print(np.shape(group_idx))
        print(np.shape(weighting_vec))
        print(N_x_bins, N_y_bins)

    try:
        if not disable_histogram:
            pixel_hist = aggregate_np(group_idx, weighting_vec, func='sum', size=(N_x_bins, N_y_bins), dtype=np.int32)
        else:
            pixel_hist = None
    except:
        raise

    return pixel_hist


def __scale_and_discretize(y_hist, N_y_bins, max_resp, min_resp, debug=False):
    """
    Normalizes and discretizes the `y_hist` array 
    
    Parameters
    ----------
    y_hist : numpy.ndarray
    N_y_bins : int
    max_resp : float
    min_resp : float
    debug : bool

    Returns
    -------
    y_hist numpy.ndarray
    """
    y_hist = y_hist.flatten()
    y_hist = np.clip(y_hist, min_resp, max_resp)
    y_hist = np.add(y_hist, -min_resp)
    y_hist = np.dot(y_hist, 1.0/(max_resp - min_resp))
    '''
    Discretize y_hist
    '''
    y_hist = np.rint(y_hist * (N_y_bins - 1))
    if debug:
        print('ymin', min(y_hist), 'ymax', max(y_hist))

    return y_hist



# -*- coding: utf-8 -*-
"""
Created on Tue Jan 05 07:55:56 2016

@author: Suhas Somnath, Chris Smith
"""

from __future__ import division, print_function, absolute_import
import h5py
import numpy as np
import sklearn.decomposition as dec

from ..io.hdf_utils import checkIfMain
from ..io.hdf_utils import getH5DsetRefs, checkAndLinkAncillary
from ..io.io_hdf5 import ioHDF5
from ..io.io_utils import check_dtype, transformToTargetType
from ..io.microdata import MicroDataGroup, MicroDataset


class Decomposition(object):
    """
    Pycroscopy wrapper around the sklearn.decomposition classes
    """

    def __init__(self, h5_main, method_name, n_components=None, *args, **kwargs):
        """
        Constructs the Decomposition object

        Parameters
        ------------
        h5_main : HDF5 dataset object
            Main dataset with ancillary spectroscopic, position indices and values datasets
        method_name : string / unicode
            Name of the sklearn.cluster estimator
        n_components : (Optional) unsigned int
            Number of components for decomposition
        *args and **kwargs : arguments to be passed to the estimator
        """

        if n_components is not None:
            kwargs['n_components'] = n_components

        allowed_methods = ['FactorAnalysis','FastICA','IncrementalPCA',
                           'MiniBatchSparsePCA','NMF','PCA','RandomizedPCA',
                           'SparsePCA','TruncatedSVD']

        # check if h5_main is a valid object - is it a hub?
        if not checkIfMain(h5_main):
            raise TypeError('Supplied dataset is not a pycroscopy main dataset')

        if method_name not in allowed_methods:
            raise TypeError('Cannot work with {} just yet'.format(method_name))

        self.h5_main = h5_main

        # Instantiate the decomposition object
        self.estimator = dec.__dict__[method_name].__call__(*args, **kwargs)
        self.method_name = method_name

        # figure out the operation that needs need to be performed to convert to real scalar
        retval = check_dtype(h5_main)
        self.data_transform_func, self.data_is_complex, self.data_is_compound, \
        self.data_n_features, self.data_n_samples, self.data_type_mult = retval

    def doDecomposition(self):
        """
        Decomposes the hdf5 dataset, and writes the ? back to the hdf5 file

        Returns
        --------
        h5_group : HDF5 Group reference
            Reference to the group that contains the decomposition results
        """
        self._fit()
        self._transform()
        return self._writeToHDF5(transformToTargetType(self.estimator.components_, self.h5_main.dtype),
                                 self.projection)

    def _fit(self):
        """
        Fits the provided dataset

        Returns
        ------
        None
        """
        # perform fit on the real dataset
        if self.method_name=='NMF':
            self.estimator.fit(self.data_transform_func(np.abs(self.h5_main)))
        else:
            self.estimator.fit(self.data_transform_func(self.h5_main))
        
    def _transform(self, data=None):
        """
        Transforms the original OR provided dataset with previously computed fit
        
        Parameters
        --------
        data : (optional) HDF5 dataset
            Dataset to apply the transform to. 
            The number of elements in the first axis of this dataset should match that of the original dataset that was fitted

        Returns
        ------
        None
        """
        if data is None:
            if self.method_name=='NMF':
                self.projection = self.estimator.transform(self.data_transform_func(np.abs(self.h5_main)))
            else:
                self.projection = self.estimator.transform(self.data_transform_func(self.h5_main))
        else:
            if isinstance(data, h5py.Dataset):
                if data.shape[0] == self.h5_main.shape[0]:
                    self.projection = self.estimator.transform(data)

    def _writeToHDF5(self, components, projection):
        """
        Writes the labels and mean response to the h5 file

        Parameters
        ------------
        labels : 1D unsigned int array
            Array of cluster labels as obtained from the fit
        mean_response : 2D numpy array
            Array of the mean response for each cluster arranged as [cluster number, response]

        Returns
        ---------
        h5_labels : HDF5 Group reference
            Reference to the group that contains the clustering results
        """
        ds_components = MicroDataset('Components', components)# equivalent to V         
        ds_projections = MicroDataset('Projection', np.float32(projection)) # equivalent of U compound        
        
        decomp_ind_mat = np.transpose(np.atleast_2d(np.arange(components.shape[0])))

        ds_decomp_inds = MicroDataset('Decomposition_Indices', np.uint32(decomp_ind_mat))
        ds_decomp_vals = MicroDataset('Decomposition_Values', np.float32(decomp_ind_mat))

        # write the labels and the mean response to h5
        decomp_slices = {'Decomp': (slice(None), slice(0, 1))}
        ds_decomp_inds.attrs['labels'] = decomp_slices
        ds_decomp_inds.attrs['units'] = ['']
        ds_decomp_vals.attrs['labels'] = decomp_slices
        ds_decomp_vals.attrs['units'] = ['']

        decomp_grp = MicroDataGroup(self.h5_main.name.split('/')[-1] + '-Decomposition_', self.h5_main.parent.name[1:])
        decomp_grp.addChildren([ds_components, ds_projections, ds_decomp_inds, ds_decomp_vals])

        decomp_grp.attrs['num_components'] = components.shape[0]
        decomp_grp.attrs['num_samples'] = self.h5_main.shape[0]
        decomp_grp.attrs['decomposition_algorithm'] = self.method_name

        '''
        Get the parameters of the estimator used and write them
        as attributes of the group
        '''
        for parm in self.estimator.get_params().keys():
            decomp_grp.attrs[parm] = self.estimator.get_params()[parm]

        hdf = ioHDF5(self.h5_main.file)
        h5_decomp_refs = hdf.writeData(decomp_grp)

        h5_components = getH5DsetRefs(['Components'], h5_decomp_refs)[0]
        h5_projections = getH5DsetRefs(['Mean_Response'], h5_decomp_refs)[0]
        h5_decomp_inds = getH5DsetRefs(['Decomposition_Indices'], h5_decomp_refs)[0]
        h5_decomp_vals = getH5DsetRefs(['Decomposition_Values'], h5_decomp_refs)[0]

        checkAndLinkAncillary(h5_projections,
                              ['Position_Indices', 'Position_Values'],
                              h5_main=self.h5_main)

        checkAndLinkAncillary(h5_projections,
                              ['Spectroscopic_Indices', 'Spectroscopic_Values'],
                              anc_refs=[h5_decomp_inds, h5_decomp_vals])

        checkAndLinkAncillary(h5_components,
                              ['Spectroscopic_Indices', 'Spectroscopic_Values'],
                              h5_main=self.h5_main)

        checkAndLinkAncillary(h5_components,
                              ['Position_Indices', 'Position_Values'],
                              anc_refs=[h5_decomp_inds, h5_decomp_vals])

        # return the h5 group object
        return h5_components.parent


# -*- coding: utf-8 -*-
"""
Created on Sun May 29 17:58:35 2016

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals

from os import path, remove  # File Path formatting
from warnings import warn

import h5py
import numpy as np  # For array operations

from .translator import Translator
from .utils import generate_dummy_main_parms, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, linkRefs
from ..io_hdf5 import ioHDF5  # Now the translator is responsible for writing the data.
from ..microdata import MicroDataGroup, MicroDataset  # building blocks for defining heirarchical storage in the H5 file


class GIVTranslator(Translator):
    """
    Translates G-mode Fast IV datasets from .mat files to .h5
    """

    def _parse_file_path(self, input_path):
        pass

    def translate(self, parm_path):
        """      
        The main function that translates the provided file into a .h5 file
        
        Parameters
        ------------
        parm_path : string / unicode
            Absolute file path of the parameters .mat file. 
            
        Returns
        ----------
        h5_path : string / unicode
            Absolute path of the translated h5 file
        """
        parm_dict, excit_wfm = self._read_parms(parm_path)
        folder_path, base_name = path.split(parm_path)
        waste, base_name = path.split(folder_path)
        
        h5_path = path.join(folder_path, base_name+'.h5')
        if path.exists(h5_path):
            remove(h5_path)

        # Now start creating datasets and populating:
        ds_spec_inds, ds_spec_vals = build_ind_val_dsets([excit_wfm.size], is_spectral=True,
                                                         labels=['Bias'], units=['V'], verbose=False)
        ds_spec_vals.data = np.atleast_2d(excit_wfm)  # The data generated above varies linearly. Override.

        ds_pos_ind, ds_pos_val = build_ind_val_dsets([parm_dict['grid_num_rows']], is_spectral=False,
                                                     steps=[1.0 * parm_dict['grid_scan_height_[m]'] /
                                                            parm_dict['grid_num_rows']],
                                                     labels=['Y'], units=['m'])
        
        # Minimize file size to the extent possible.
        # DAQs are rated at 16 bit so float16 should be most appropriate.
        # For some reason, compression is effective only on time series data
        ds_raw_data = MicroDataset('Raw_Data', data=[],
                                   maxshape=(parm_dict['grid_num_rows'], excit_wfm.size),
                                   dtype=np.float16, chunking=(1, excit_wfm.size), compression='gzip')
        ds_raw_data.attrs['quantity'] = ['Current']
        ds_raw_data.attrs['units'] = ['1E-{} A'.format(parm_dict['IO_amplifier_gain'])]
        
        aux_ds_names = ['Position_Indices', 'Position_Values',
                        'Spectroscopic_Indices', 'Spectroscopic_Values']
        
        # Until a better method is provided....        
        h5_f = h5py.File(path.join(folder_path, 'line_1.mat'), 'r')
        num_ai_chans = h5_f['data'].shape[1]
        h5_f.close()        
                
        # technically should change the date, etc.              
        spm_data = MicroDataGroup('')
        global_parms = generate_dummy_main_parms()
        global_parms['data_type'] = 'gIV'
        global_parms['translator'] = 'gIV'
        spm_data.attrs = global_parms
        meas_grp = MicroDataGroup('Measurement_000') 
        spm_data.addChildren([meas_grp])
        
        hdf = ioHDF5(h5_path)
        # spm_data.showTree()
        hdf.writeData(spm_data, print_log=False)

        self.raw_datasets = list()
        
        for chan_index in range(num_ai_chans):
            
            chan_grp = MicroDataGroup('{:s}{:03d}'.format('Channel_', chan_index), '/Measurement_000/')
            chan_grp.attrs = parm_dict
            chan_grp.addChildren([ds_pos_ind, ds_pos_val, ds_spec_inds, ds_spec_vals,
                                  ds_raw_data])
            h5_refs = hdf.writeData(chan_grp, print_log=False)
            h5_raw = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
            linkRefs(h5_raw, getH5DsetRefs(aux_ds_names, h5_refs))
            self.raw_datasets.append(h5_raw)
            
        # Now that the N channels have been made, populate them with the actual data....
        self._read_data(parm_dict, folder_path)

        hdf.close()        
        return h5_path

    def _read_data(self, parm_dict, folder_path):
        """
        Reads raw data and populates the h5 datasets

        Parameters
        ----------
        parm_dict : Dictionary
            dictionary containing parameters for this data
        folder_path : string / unicode
            Absolute path of folder containing the data
        """
        if parm_dict['excitation_extra_pts'] == 0:
            main_data = slice(parm_dict['excitation_pulse_points'], None)
        else:
            main_data = slice(parm_dict['excitation_pulse_points'], -1 * parm_dict['excitation_extra_pts'])

        for line_ind in range(parm_dict['grid_num_rows']):
            if line_ind % np.round(parm_dict['grid_num_rows']/10) == 0:
                print('Reading data in line {} of {}'.format(line_ind+1, parm_dict['grid_num_rows']))
            file_path = path.join(folder_path, 'line_'+str(line_ind+1)+'.mat')
            if path.exists(file_path):
                h5_f = h5py.File(file_path, 'r')
                h5_data = h5_f['data']
                if h5_data.shape[0] >= parm_dict['excitation_length'] and h5_data.shape[1] == len(self.raw_datasets):
                    for chan, h5_chan in enumerate(self.raw_datasets):
                        h5_chan[line_ind, :] = np.float16(h5_data[main_data, chan])
                        h5_chan.file.flush()
                else:
                    warn('No data found for Line '+str(line_ind))
            else:
                warn('File not found for: line '+str(line_ind))
        print('Finished reading all data!')

    @staticmethod
    def _read_parms(parm_path):
        """
        Copies experimental parameters from the .mat file to a dictionary
        
        Parameters
        ----------
        parm_path : string / unicode
            Absolute path of the parameters file
        
        Returns 
        -------
        parm_dict : dictionary
            Dictionary containing all relevant parameters
        excit_wfm : 1d numpy float array
            Excitation waveform
        """
        h5_f = h5py.File(parm_path, 'r')
        parm_dict = dict()
        
        parm_dict['IO_samp_rate_[Hz]'] = np.uint32(h5_f['samp_rate'][0][0])
        parm_dict['IO_amplifier_gain'] = np.uint32(h5_f['amp_gain'][0][0])

        parm_dict['excitation_frequency_[Hz]'] = np.float32(h5_f['frequency'][0][0])
        parm_dict['excitation_amplitude_[V]'] = np.float32(h5_f['amplitude'][0][0])
        parm_dict['excitation_offset_[V]'] = np.float32(h5_f['offset'][0][0])
        excit_wfm = np.float32(np.squeeze(h5_f['excit_wfm'].value))
        
        # Make sure to truncate the data to the point when the 
        pts_per_cycle = int(np.round(1.0*parm_dict['IO_samp_rate_[Hz]']/parm_dict['excitation_frequency_[Hz]']))
        extra_pts = len(excit_wfm) % pts_per_cycle
        parm_dict['excitation_extra_pts'] = extra_pts

        """ New versions could have used a pulse at the beginning of the line to "wake up" the material.
        This pulse is always an integer number of cycles long
        For now, let's remove this section from the excitation waveform and data"""

        # This pulse may or may not have been used:
        try:
            pulse_duration = np.float32(h5_f['pulse_duration'][0][0])
            pulse_height = np.float32(h5_f['pulse_height'][0][0])
        except KeyError:
            pulse_duration = 0.0
            pulse_height = 0.0
        if pulse_duration == 0.0:
            pulse_height = 0.0

        parm_dict['excitation_pulse_height_[V]'] = np.float32(pulse_height)
        parm_dict['excitation_pulse_time_[sec]'] = np.float32(pulse_duration)
        pulse_points = int(np.round(pulse_duration * parm_dict['IO_samp_rate_[Hz]']))
        parm_dict['excitation_pulse_points'] = np.uint32(pulse_points)

        line_time = np.float32(h5_f['line_time'][0][0]) - pulse_duration
        excess_time = line_time - 1.0*extra_pts/parm_dict['IO_samp_rate_[Hz]']
        parm_dict['excitation_duration_[sec]'] = line_time - excess_time
        if extra_pts > 0:
            excit_wfm = excit_wfm[pulse_points:-extra_pts]
        else:
            excit_wfm = excit_wfm[pulse_points:]
        parm_dict['excitation_length'] = len(excit_wfm)

        parm_dict['grid_num_rows'] = np.uint32(h5_f['num_lines'][0][0])
        parm_dict['grid_num_cols'] = np.uint32(np.floor(len(excit_wfm) / pts_per_cycle))
        
        parm_dict['grid_scan_height_[m]'] = np.float32(h5_f['scan_height'][0][0])
        parm_dict['grid_scan_width_[m]'] = np.float32(h5_f['scan_width'][0][0])
        parm_dict['grid_scan_speed_[ms-1]'] = np.float32(h5_f['scan_speed'][0][0])

        h5_f.close()
        return parm_dict, excit_wfm
